import asyncio
import datetime
import importlib
import logging
import re
import tempfile
import unittest.mock
from copy import copy
from pathlib import Path
from typing import Any, Dict, Hashable, Optional
from xml.etree import ElementTree as ET

from slixmpp import JID, Message, Presence
from slixmpp import __version__ as slix_version
from slixmpp.exceptions import XMPPError
from slixmpp.plugins.xep_0030 import DiscoInfo
from slixmpp.plugins.xep_0082 import format_datetime
from slixmpp.plugins.xep_0356.permissions import (
    MessagePermission,
    Permissions,
    PresencePermission,
    RosterAccess,
)
from sqlalchemy import delete

from slidge import *
from slidge import (
    FormField,
    GatewayUser,
    LegacyBookmarks,
    LegacyContact,
    LegacyRoster,
    SearchResult,
)
from slidge.command.categories import ADMINISTRATION
from slidge.core import config
from slidge.core.mixins.attachment import AttachmentMixin
from slidge.core.session import _sessions
from slidge.db.models import Contact
from slidge.util.test import SlidgeTest
from slidge.util.types import LegacyAttachment, LegacyContactType, LegacyMessageType

received_presences: list[Optional[Presence]] = []
text_received_by_juliet = []
composing_chat_states_received_by_juliet = []
unregistered = []
reactions_received_by_juliet = []


class Gateway(BaseGateway):
    COMPONENT_NAME = "SLIDGE TEST"

    SEARCH_FIELDS = [FormField(var="leg", label="Enter the legacy ID")]
    SEARCH_TITLE = "Search for legacy contacts"

    GROUPS = True

    LEGACY_CONTACT_ID_TYPE = int
    LEGACY_MSG_ID_TYPE = int

    async def unregister(self, user: GatewayUser):
        unregistered.append(user)


class Session(BaseSession):
    @staticmethod
    def xmpp_to_legacy_msg_id(i: str):
        return int(i)

    async def on_search(self, form_values: Dict[str, str]):
        if form_values["leg"] == "exists":
            return SearchResult(
                fields=[FormField(var="jid", label="JID", type="jid-single")],
                items=[{"jid": "exists@example.com"}],
            )

    def __init__(self, user):
        super().__init__(user)

    async def wait_for_ready(self, timeout=10):
        return

    async def login(self):
        pass

    async def logout(self):
        pass

    async def on_text(
        self,
        chat: LegacyContact,
        text: str,
        *,
        reply_to=None,
        reply_to_msg_id=None,
        reply_to_fallback_text: Optional[str] = None,
        thread=None,
        link_previews=(),
    ):
        if chat.jid.node == "juliet":
            text_received_by_juliet.append((text, chat))
        assert self.user_jid.bare == "romeo@montague.lit"
        assert self.user_jid == JID("romeo@montague.lit")
        chat.send_text("I love you")
        return 0

    async def on_composing(self, c: LegacyContact, thread=None):
        composing_chat_states_received_by_juliet.append(c)

    async def on_displayed(
        self, c: LegacyContact, legacy_msg_id: Hashable, thread=None
    ):
        pass

    async def on_react(
        self,
        c: LegacyContact,
        legacy_msg_id: LegacyMessageType,
        emojis: list[str],
        thread=None,
    ):
        if c.jid.node == "juliet":
            for e in emojis:
                reactions_received_by_juliet.append([legacy_msg_id, e])


class Roster(LegacyRoster):
    @staticmethod
    async def jid_username_to_legacy_id(jid_username: str) -> int:
        log.debug("Requested JID to legacy: %s", jid_username)
        if jid_username == "juliet":
            return 123
        elif jid_username == "new-friend":
            return 456
        else:
            raise XMPPError(text="Only juliet", condition="item-not-found")

    @staticmethod
    async def legacy_id_to_jid_username(legacy_id: int) -> str:
        if legacy_id == 123:
            return "juliet"
        elif legacy_id == 456:
            return "new-friend"
        else:
            raise RuntimeError


class Bookmarks(LegacyBookmarks):
    @staticmethod
    async def jid_local_part_to_legacy_id(local_part):
        if local_part != "room":
            raise XMPPError("item-not-found")
        else:
            return local_part

    async def fill(self):
        await self.by_legacy_id("room1")
        await self.by_legacy_id("room2")


class Base(SlidgeTest):
    plugin = globals()

    def setUp(self):
        super().setUp()
        with self.xmpp.store.session() as orm:
            user = GatewayUser(
                jid=JID("romeo@montague.lit/gajim").bare,
                legacy_module_data={"username": "romeo", "city": ""},
                preferences={"sync_avatar": True, "sync_presence": True},
            )
            orm.add(user)
            orm.commit()
        self.get_romeo_session().logged = True

    @staticmethod
    def get_romeo_session() -> Session:
        return BaseSession.get_self_or_unique_subclass().from_jid(
            JID("romeo@montague.lit")
        )

    @property
    def juliet(self) -> LegacyContact:
        session = BaseSession.get_self_or_unique_subclass().from_jid(
            JID("romeo@montague.lit")
        )
        return self.run_coro(session.contacts.by_jid(JID("juliet@aim.shakespeare.lit")))


class TestAimShakespeareBase(Base):
    def loop(self, x):
        self.run_coro(x)

    def test_jabber_iq_gateway(self):
        self.recv(  # language=XML
            """
            <iq type='get'
                to='aim.shakespeare.lit'
                from='romeo@montague.lit'
                id='gate1'>
              <query xmlns='jabber:iq:gateway' />
            </iq>
            """
        )
        self.send(  # language=XML
            f"""
            <iq type='result'
                from='aim.shakespeare.lit'
                to='romeo@montague.lit'
                id='gate1'>
              <query xmlns='jabber:iq:gateway'>
                <desc>{Gateway.SEARCH_TITLE}</desc>
                <prompt>{Gateway.SEARCH_FIELDS[0].label}</prompt>
              </query>
            </iq>
            """
        )
        self.recv(  # language=XML
            """
            <iq type='set'
                to='aim.shakespeare.lit'
                from='romeo@montague.lit'
                id='gate1'>
              <query xmlns='jabber:iq:gateway'>
                <prompt>exists</prompt>
              </query>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq type='result'
                from='aim.shakespeare.lit'
                to='romeo@montague.lit'
                id='gate1'>
              <query xmlns='jabber:iq:gateway'>
                <jid>exists@example.com</jid>
              </query>
            </iq>
            """
        )
        self.recv(  # language=XML
            """
            <iq type='set'
                to='aim.shakespeare.lit'
                from='romeo@montague.lit'
                id='gate1'>
              <query xmlns='jabber:iq:gateway'>
                <prompt>not-exists</prompt>
              </query>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq type='error'
                from='aim.shakespeare.lit'
                to='romeo@montague.lit'
                id='gate1'>
              <error type="cancel">
                <item-not-found xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">No contact was found with the info you provided.</text>
              </error>
            </iq>
            """,
            use_values=False,
        )

    def test_jabber_iq_gateway_on_contact(self):
        self.recv(  # language=XML
            """
            <iq type='get'
                to='juliet@aim.shakespeare.lit'
                from='romeo@montague.lit'
                id='gate1'>
              <query xmlns='jabber:iq:gateway' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq type='error'
                from='juliet@aim.shakespeare.lit'
                to='romeo@montague.lit'
                id='gate1'>
              <error type="modify">
                <bad-request xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">This can only be used on the component JID</text>
              </error>
            </iq>
            """
        )

    def test_from_romeo_to_eve(self):
        self.recv(  # language=XML
            """
            <message type='chat'
                     to='eve@aim.shakespeare.lit'
                     from='romeo@montague.lit'>
              <body>Art thou not Romeo, and a Montague?</body>
            </message>
            """
        )
        self.send(  # language=XML
            """
            <message type="error"
                     to="romeo@montague.lit"
                     from="eve@aim.shakespeare.lit">
              <error type="cancel">
                <item-not-found xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">Only juliet</text>
              </error>
            </message>
            """,
            use_values=False,
        )

    def test_from_romeo_to_juliet(self):
        self.recv(  # language=XML
            """
            <message type='chat'
                     to='juliet@aim.shakespeare.lit'
                     from='romeo@montague.lit'>
              <body>Art thou not Romeo, and a Montague?</body>
            </message>
            """
        )
        text, contact = text_received_by_juliet[-1]
        assert text == "Art thou not Romeo, and a Montague?"
        assert contact.legacy_id == 123
        m: Message = self.next_sent()
        assert m.get_from() == "juliet@aim.shakespeare.lit/slidge"
        assert m["body"] == "I love you"
        m2 = copy(
            m
        )  # there must be a better way to check for the presence of the markable thing
        m2.enable("markable")
        assert m == m2
        text_received_by_juliet.clear()

    def test_delivery_receipt(self):
        self.xmpp.PROPER_RECEIPTS = True
        self.recv(  # language=XML
            """
            <message type='chat'
                     to='juliet@aim.shakespeare.lit/slidge'
                     from='romeo@montague.lit/prout'
                     id="123">
              <body>Art thou not Romeo, and a Montague?</body>
              <request xmlns='urn:xmpp:receipts' />
            </message>
            """
        )
        self.next_sent()  # auto reply in our test plugin
        session = BaseSession.get_self_or_unique_subclass().from_jid(
            JID("romeo@montague.lit")
        )
        juliet = self.run_coro(
            session.contacts.by_jid(JID("juliet@aim.shakespeare.lit"))
        )
        juliet.received("123")
        self.send(  # language=XML
            """
            <message xmlns="jabber:component:accept"
                     type="chat"
                     to="romeo@montague.lit"
                     from="juliet@aim.shakespeare.lit/slidge">
              <received xmlns="urn:xmpp:receipts"
                        id="123" />
            </message>
            """
        )
        self.send(  # language=XML
            """
            <message xmlns="jabber:component:accept"
                     type="chat"
                     to="romeo@montague.lit"
                     from="juliet@aim.shakespeare.lit/slidge">
              <received xmlns="urn:xmpp:chat-markers:0"
                        id="123" />
            </message>
            """
        )
        assert self.next_sent() is None

    def test_romeo_composing(self):
        self.recv(  # language=XML
            """
            <message type='chat'
                     to='juliet@aim.shakespeare.lit'
                     from='romeo@montague.lit'>
              <composing xmlns='http://jabber.org/protocol/chatstates' />
            </message>
            """
        )
        assert len(composing_chat_states_received_by_juliet) == 1
        assert composing_chat_states_received_by_juliet[0].legacy_id == 123

    def test_from_eve_to_juliet(self):
        # just ignore messages from unregistered users
        self.recv(  # language=XML
            """
            <message type='chat'
                     from='eve@aim.shakespeare.lit'
                     to='juliet@montague.lit'>
              <body>Art thou not Romeo, and a Montague?</body>
            </message>
            """
        )
        self.send(  # language=XML
            """
            <message type="error"
                     from="juliet@montague.lit"
                     to="eve@aim.shakespeare.lit">
              <error type="auth">
                <registration-required xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">You are not registered to this gateway</text>
              </error>
            </message>
            """
        )

    def test_juliet_sends_text(self):
        session = BaseSession.get_self_or_unique_subclass().from_jid(
            JID("romeo@montague.lit")
        )
        juliet = self.run_coro(
            session.contacts.by_jid(JID("juliet@aim.shakespeare.lit"))
        )
        juliet.send_text(body="What what?")

        msg = self.next_sent()

        assert msg["from"] == f"juliet@aim.shakespeare.lit/{LegacyContact.RESOURCE}"
        assert msg["to"] == "romeo@montague.lit"
        assert msg["body"] == "What what?"

    def test_unregister(self):
        assert len(unregistered) == 0
        self.recv(  # language=XML
            """
            <message type='chat'
                     to='juliet@aim.shakespeare.lit'
                     from='romeo@montague.lit'>
              <composing xmlns='http://jabber.org/protocol/chatstates' />
            </message>
            """
        )  # this creates a session
        self.recv(  # language=XML
            """
            <iq from='romeo@montague.lit'
                type='set'
                to='aim.shakespeare.lit'>
              <query xmlns='jabber:iq:register'>
                <remove />
              </query>
            </iq>
            """
        )
        assert len(unregistered) == 1
        assert unregistered[0].user.jid == "romeo@montague.lit"
        unregistered.clear()

    def test_jid_validator(self):
        self.xmpp.jid_validator = re.compile(".*@noteverybody")
        self.recv(  # language=XML
            f"""
            <iq from='eve@nothingshakespearian'
                type='get'
                to='aim.shakespeare.lit'
                id="0">
              <query xmlns='jabber:iq:register'></query>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                from="aim.shakespeare.lit"
                type="error"
                to="eve@nothingshakespearian"
                id="0">
              <error type="cancel">
                <not-allowed xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">Your account is not allowed to use this gateway. The admin controls that with the USER_JID_VALIDATOR option.</text>
              </error>
            </iq>
            """,
            use_values=False,
        )
        self.recv(  # language=XML
            """
            <iq from='eve@nothingshakespearian'
                type='set'
                to='aim.shakespeare.lit'
                id='1'>
              <query xmlns='jabber:iq:register'>
                <username>bill</username>
                <password>Calliope</password>
              </query>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                from="aim.shakespeare.lit"
                type="error"
                to="eve@nothingshakespearian"
                id="1">
              <error type="cancel">
                <not-allowed xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">Your account is not allowed to use this gateway. The admin controls that with the USER_JID_VALIDATOR option.</text>
              </error>
            </iq>
            """,
            use_values=False,
        )
        self.xmpp.jid_validator = re.compile(".*")

    def test_reactions(self):
        self.recv(  # language=XML
            """
            <message type='chat'
                     to='juliet@aim.shakespeare.lit'
                     from='romeo@montague.lit'>
              <reactions id='555'
                         xmlns='urn:xmpp:reactions:0'>
                <reaction>👋</reaction>
                <reaction>🐢</reaction>
              </reactions>
            </message>
            """
        )
        assert len(reactions_received_by_juliet) == 2
        msg_id, emoji = reactions_received_by_juliet[0]
        assert msg_id == 555
        assert emoji == "👋"
        msg_id, emoji = reactions_received_by_juliet[1]
        assert msg_id == 555
        assert emoji == "🐢"

        session = BaseSession.get_self_or_unique_subclass().from_jid(
            JID("romeo@montague.lit")
        )
        juliet = self.run_coro(
            session.contacts.by_jid(JID("juliet@aim.shakespeare.lit"))
        )
        juliet.react("legacy1", ["👋", "👋"])
        msg = self.next_sent()
        assert msg["reactions"]["id"] == "legacy1"
        assert len(msg["reactions"]) == 1
        for r in msg["reactions"]:
            assert r["value"] == "👋"
        reactions_received_by_juliet.clear()

    def test_reactions_fallback(self):
        self.recv(  # language=XML
            """
            <message type='chat'
                     to='juliet@aim.shakespeare.lit'
                     from='romeo@montague.lit'>
              <body>&gt; huuuuu\n👍</body>
              <request xmlns="urn:xmpp:receipts" />
              <markable xmlns="urn:xmpp:chat-markers:0" />
              <origin-id xmlns="urn:xmpp:sid:0"
                         id="e05cadc7-b752-4ed9-b47e-5ad862a4430d" />
              <reply xmlns="urn:xmpp:reply:0"
                     to="5308963142@telegram.slidge.im/slidge"
                     id="4940890112" />
              <fallback xmlns="urn:xmpp:fallback:0"
                        for="urn:xmpp:reply:0">
                <body end="9"
                      start="0" />
              </fallback>
              <fallback xmlns="urn:xmpp:fallback:0"
                        for="urn:xmpp:reactions:0">
                <body />
              </fallback>
              <reactions xmlns="urn:xmpp:reactions:0"
                         id="4940890112">
                <reaction>👍</reaction>
              </reactions>
              <thread>0516cd85-4d78-4d29-bc87-378e33d820b3</thread>
              <active xmlns="http://jabber.org/protocol/chatstates" />
            </message>
            """
        )
        assert len(reactions_received_by_juliet) == 1
        msg_id, emoji = reactions_received_by_juliet[0]
        assert msg_id == 4940890112
        assert emoji == "👍"
        assert len(text_received_by_juliet) == 0
        text_received_by_juliet.clear()
        reactions_received_by_juliet.clear()

    def test_last_seen(self):
        session = BaseSession.get_self_or_unique_subclass().from_jid(
            JID("romeo@montague.lit")
        )
        juliet = self.run_coro(
            session.contacts.by_jid(JID("juliet@aim.shakespeare.lit"))
        )
        juliet.is_friend = True
        now = datetime.datetime.now(datetime.timezone.utc)
        juliet.away(last_seen=now)
        sent = self.next_sent()
        assert sent["idle"]["since"] == now

    def test_disco_adhoc_commands_unregistered(self):
        self.recv(  # language=XML
            f"""
            <iq type='get'
                from='requester@domain'
                to='{self.xmpp.boundjid.bare}'>
              <query xmlns='http://jabber.org/protocol/disco#items'
                     node='http://jabber.org/protocol/commands' />
            </iq>
            """
        )
        self.send(  # language=XML
            f"""
            <iq type='result'
                to='requester@domain'
                from='{self.xmpp.boundjid.bare}'
                id='1'>
              <query xmlns='http://jabber.org/protocol/disco#items'
                     node='http://jabber.org/protocol/commands'>
                <item jid="aim.shakespeare.lit"
                      node="https://slidge.im/command/core/admin"
                      name="🛷️ Slidge administration" />
                <item jid="aim.shakespeare.lit"
                      node="jabber:iq:register"
                      name="📝 Register to the gateway" />
              </query>
            </iq>
            """
        )

    def test_disco_adhoc_commands_as_logged_user(self):
        self.recv(  # language=XML
            f"""
            <iq type='get'
                from='romeo@montague.lit/gajim'
                to='{self.xmpp.boundjid.bare}'>
              <query xmlns='http://jabber.org/protocol/disco#items'
                     node='http://jabber.org/protocol/commands' />
            </iq>
            """
        )
        self.send(  # language=XML
            f"""
            <iq type='result'
                to='romeo@montague.lit/gajim'
                from='{self.xmpp.boundjid.bare}'
                id='1'>
              <query xmlns='http://jabber.org/protocol/disco#items'
                     node='http://jabber.org/protocol/commands'>
                <item jid="aim.shakespeare.lit"
                      name="🛷️ Slidge administration"
                      node="https://slidge.im/command/core/admin" />
                <item jid="aim.shakespeare.lit"
                      node="https://slidge.im/command/core/contacts"
                      name="👤 Contacts" />
                <item jid="aim.shakespeare.lit"
                      node="https://slidge.im/command/core/groups"
                      name="👥 Groups" />
                <item jid="aim.shakespeare.lit"
                      node="https://slidge.im/command/core/preferences"
                      name="⚙️ Preferences" />
                <item jid="aim.shakespeare.lit"
                      node="https://slidge.im/command/core/unregister"
                      name="❌ Unregister from the gateway" />
              </query>
            </iq>
            """
        )

    def test_disco_adhoc_commands_as_non_logged_user(self):
        self.get_romeo_session().logged = False
        self.recv(  # language=XML
            f"""
            <iq type='get'
                from='romeo@montague.lit/gajim'
                to='{self.xmpp.boundjid.bare}'>
              <query xmlns='http://jabber.org/protocol/disco#items'
                     node='http://jabber.org/protocol/commands' />
            </iq>
            """
        )
        self.send(  # language=XML
            f"""
            <iq type='result'
                to='romeo@montague.lit/gajim'
                from='{self.xmpp.boundjid.bare}'
                id='1'>
              <query xmlns='http://jabber.org/protocol/disco#items'
                     node='http://jabber.org/protocol/commands'>
                <item jid="aim.shakespeare.lit"
                      node="https://slidge.im/command/core/admin"
                      name="🛷️ Slidge administration" />
                <item jid="aim.shakespeare.lit"
                      node="https://slidge.im/command/core/re-login"
                      name="🔐 Re-login to the legacy network" />
                <item jid="aim.shakespeare.lit"
                      node="https://slidge.im/command/core/preferences"
                      name="⚙️ Preferences" />
                <item jid="aim.shakespeare.lit"
                      node="https://slidge.im/command/core/unregister"
                      name="❌ Unregister from the gateway" />
              </query>
            </iq>
            """
        )
        self.get_romeo_session().logged = True

    def test_disco_adhoc_commands_as_admin(self):
        # monkeypatch.setattr(config, "ADMINS", ("romeo@montague.lit",))
        config.ADMINS = (JID("admin@montague.lit"),)
        self.recv(  # language=XML
            f"""
            <iq type='get'
                from='admin@montague.lit/gajim'
                to='{self.xmpp.boundjid.bare}'>
              <query xmlns='http://jabber.org/protocol/disco#items'
                     node='http://jabber.org/protocol/commands' />
            </iq>
            """
        )
        self.send(  # language=XML
            f"""
            <iq type='result'
                to='admin@montague.lit/gajim'
                from='{self.xmpp.boundjid.bare}'
                id='1'>
              <query xmlns='http://jabber.org/protocol/disco#items'
                     node='http://jabber.org/protocol/commands'>
                <item jid="aim.shakespeare.lit"
                      node="{ADMINISTRATION.node}"
                      name="{ADMINISTRATION.name}" />
                <item jid="aim.shakespeare.lit"
                      node="jabber:iq:register"
                      name="📝 Register to the gateway" />
              </query>
            </iq>
            """
        )
        config.ADMINS = ()

    def test_disco_adhoc_commands_as_unauthorized(self):
        self.xmpp.jid_validator = re.compile("nothing")
        self.recv(  # language=XML
            f"""
            <iq type='get'
                from='admin@forbidden.lit/gajim'
                to='{self.xmpp.boundjid.bare}'>
              <query xmlns='http://jabber.org/protocol/disco#items'
                     node='http://jabber.org/protocol/commands' />
            </iq>
            """
        )
        st = self.next_sent()
        assert st["error"]["condition"] == "forbidden"
        self.xmpp.jid_validator = re.compile(".*")

    def test_disco_adhoc_command_register(self):
        self.xmpp.jid_validator = re.compile(".*@whatever.com")
        self.recv(  # language=XML
            f"""
            <iq type='get'
                from='candidate@whatever.com'
                to='{self.xmpp.boundjid.bare}'>
              <query xmlns='http://jabber.org/protocol/disco#items'
                     node='http://jabber.org/protocol/commands' />
            </iq>
            """
        )
        self.send(  # language=XML
            f"""
            <iq type='result'
                to='candidate@whatever.com'
                from='{self.xmpp.boundjid.bare}'
                id='1'>
              <query xmlns='http://jabber.org/protocol/disco#items'
                     node='http://jabber.org/protocol/commands'>
                <item jid="aim.shakespeare.lit"
                      node="{ADMINISTRATION.node}"
                      name="{ADMINISTRATION.name}" />
                <item jid="aim.shakespeare.lit"
                      node="jabber:iq:register"
                      name="📝 Register to the gateway" />
              </query>
            </iq>
            """
        )
        self.xmpp.jid_validator = re.compile(".*")

    def test_adhoc_forbidden_non_admin(self):
        with unittest.mock.patch(
            "slixmpp.plugins.xep_0050.adhoc.XEP_0050.new_session",
            return_value="session-id",
        ):
            self.recv(  # language=XML
                f"""
            <iq type="set"
                from="test@localhost/gajim"
                to="{self.xmpp.boundjid.bare}"
                id="123">
              <command xmlns="http://jabber.org/protocol/commands"
                       action="execute"
                       node="{ADMINISTRATION.node}" />
            </iq>
            """
            )
        self.send(  # language=XML
            """
            <iq type="result"
                from="aim.shakespeare.lit"
                to="test@localhost/gajim"
                id="123">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="https://slidge.im/command/core/admin"
                       sessionid="session-id"
                       status="executing">
                <actions>
                  <next />
                </actions>
                <x xmlns="jabber:x:data"
                   type="form">
                  <title>🛷️ Slidge administration</title>
                  <field var="command"
                         label="Command"
                         type="list-single">
                    <option label="ℹ️ Server information">
                      <value>https://slidge.im/command/core/admin/info</value>
                    </option>
                    <value />
                  </field>
                </x>
              </command>
            </iq>
            """,
            use_values=False,
        )

    def test_disco_component(self):
        self.recv(  # language=XML
            f"""
            <iq type="get"
                from="test@localhost/gajim"
                to="{self.xmpp.boundjid.bare}"
                id="123">
              <query xmlns='http://jabber.org/protocol/disco#info' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="aim.shakespeare.lit"
                to="test@localhost/gajim"
                id="123">
              <query xmlns="http://jabber.org/protocol/disco#info">
                <identity category="conference"
                          type="text"
                          name="SLIDGE TEST" />
                <identity category="account"
                          type="registered"
                          name="SLIDGE TEST" />
                <identity category="pubsub"
                          type="pep"
                          name="SLIDGE TEST" />
                <identity category="gateway"
                          type=""
                          name="SLIDGE TEST" />
                <feature var="jabber:iq:search" />
                <feature var="jabber:iq:register" />
                <feature var="jabber:iq:gateway" />
                <feature var="urn:ietf:params:xml:ns:vcard-4.0" />
                <feature var="http://jabber.org/protocol/pubsub#event" />
                <feature var="http://jabber.org/protocol/pubsub#retrieve-items" />
                <feature var="http://jabber.org/protocol/pubsub#persistent-items" />
                <feature var="http://jabber.org/protocol/muc" />
                <feature var="http://jabber.org/protocol/commands" />
                <feature var="urn:xmpp:mam:2" />
                <feature var="urn:xmpp:mam:2#extended" />
                <feature var="urn:xmpp:ping" />
                <feature var="urn:xmpp:occupant-id:0" />
                <feature var="urn:xmpp:avatar:metadata+notify" />
                <feature var="urn:xmpp:hats:0" />
                <feature var="urn:xmpp:mds:displayed:0+notify" />
                <feature var="urn:xmpp:mds:displayed:0" />
                <feature var='urn:xmpp:chat-markers:0' />
              </query>
            </iq>
            """
        )

    def test_disco_local_part_unregistered(self):
        self.recv(  # language=XML
            f"""
            <iq type="get"
                from="test@localhost/gajim"
                to="juliet@{self.xmpp.boundjid.bare}"
                id="123">
              <query xmlns='http://jabber.org/protocol/disco#info' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="error"
                from="juliet@aim.shakespeare.lit"
                to="test@localhost/gajim"
                id="123">
              <error type="auth">
                <registration-required xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
              </error>
            </iq>
            """,
            use_values=False,
        )

    def test_disco_registered_existing_contact(self):
        self.recv(  # language=XML
            f"""
            <iq type="get"
                from="romeo@montague.lit/gajim"
                to="juliet@{self.xmpp.boundjid.bare}/slidge"
                id="123">
              <query xmlns='http://jabber.org/protocol/disco#info' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="juliet@aim.shakespeare.lit/slidge"
                to="romeo@montague.lit/gajim"
                id="123">
              <query xmlns="http://jabber.org/protocol/disco#info">
                <identity category="client"
                          type="pc"
                          name="SLIDGE TEST" />
                <feature var="http://jabber.org/protocol/chatstates" />
                <feature var="urn:xmpp:receipts" />
                <feature var="urn:xmpp:message-correct:0" />
                <feature var="urn:xmpp:chat-markers:0" />
                <feature var="jabber:x:oob" />
                <feature var="urn:xmpp:reactions:0" />
                <feature var="urn:xmpp:message-retract:0" />
                <feature var="urn:xmpp:reply:0" />
                <feature var="urn:ietf:params:xml:ns:vcard-4.0" />
                <feature var='jabber:x:conference' />
              </query>
            </iq>
            """,
        )

    def test_disco_registered_existing_contact_bare_jid(self):
        self.recv(  # language=XML
            f"""
            <iq type="get"
                from="romeo@montague.lit/gajim"
                to="juliet@{self.xmpp.boundjid.bare}"
                id="123">
              <query xmlns='http://jabber.org/protocol/disco#info' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="juliet@aim.shakespeare.lit"
                to="romeo@montague.lit/gajim"
                id="123">
              <query xmlns="http://jabber.org/protocol/disco#info">
                <identity category="account"
                          type="registered"
                          name="SLIDGE TEST" />
                <identity category="pubsub"
                          type="pep"
                          name="SLIDGE TEST" />
                <feature var="http://jabber.org/protocol/pubsub" />
                <feature var="http://jabber.org/protocol/pubsub#retrieve-items" />
                <feature var="http://jabber.org/protocol/pubsub#subscribe" />
              </query>
            </iq>
            """,
        )

    def test_disco_items_registered_existing_contact(self):
        session = BaseSession.get_self_or_unique_subclass().from_jid(
            JID("romeo@montague.lit")
        )
        self.run_coro(session.bookmarks.fill())
        self.recv(  # language=XML
            f"""
            <iq type="get"
                from="romeo@montague.lit/gajim"
                to="juliet@{self.xmpp.boundjid.bare}"
                id="123">
              <query xmlns='http://jabber.org/protocol/disco#items'
                     node="http://jabber.org/protocol/commands" />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="juliet@aim.shakespeare.lit"
                to="romeo@montague.lit/gajim"
                id="123">
              <query xmlns="http://jabber.org/protocol/disco#items" />
            </iq>
            """,
        )

    def test_disco_restricted_reaction(self):
        session = BaseSession.get_self_or_unique_subclass().from_jid(
            JID("romeo@montague.lit")
        )
        juliet: LegacyContact = self.run_coro(
            session.contacts.by_jid(JID("juliet@aim.shakespeare.lit"))
        )
        LegacyContact.REACTIONS_SINGLE_EMOJI = True
        self.recv(  # language=XML
            f"""
            <iq type="get"
                from="romeo@montague.lit/gajim"
                to="juliet@{self.xmpp.boundjid.bare}/slidge"
                id="123">
              <query xmlns='http://jabber.org/protocol/disco#info' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="juliet@aim.shakespeare.lit/slidge"
                to="romeo@montague.lit/gajim"
                id="123">
              <query xmlns="http://jabber.org/protocol/disco#info">
                <identity category="client"
                          type="pc"
                          name="SLIDGE TEST" />
                <feature var="http://jabber.org/protocol/chatstates" />
                <feature var="urn:xmpp:receipts" />
                <feature var="urn:xmpp:message-correct:0" />
                <feature var="urn:xmpp:chat-markers:0" />
                <feature var="jabber:x:oob" />
                <feature var="urn:xmpp:reactions:0" />
                <feature var="urn:xmpp:message-retract:0" />
                <feature var="urn:xmpp:reply:0" />
                <feature var="urn:ietf:params:xml:ns:vcard-4.0" />
                <feature var="jabber:x:conference" />
                <x xmlns='jabber:x:data'
                   type='result'>
                  <field var='FORM_TYPE'
                         type='hidden'>
                    <value>urn:xmpp:reactions:0:restrictions</value>
                  </field>
                  <field var='max_reactions_per_user'
                         type='number'>
                    <value>1</value>
                  </field>
                </x>
              </query>
            </iq>
            """,
        )
        LegacyContact.REACTIONS_SINGLE_EMOJI = False

    def test_non_existing_contact(self):
        self.recv(  # language=XML
            f"""
            <message from="romeo@montague.lit/gajim"
                     to="nope@{self.xmpp.boundjid.bare}/slidge"
                     id="123">
              <body>DSAD</body>
            </message>
            """
        )
        self.send(  # language=XML
            """
            <message xmlns="jabber:component:accept"
                     type="error"
                     from="nope@aim.shakespeare.lit/slidge"
                     to="romeo@montague.lit/gajim"
                     id="123">
              <error type="cancel">
                <item-not-found xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">Only juliet</text>
              </error>
            </message>
            """,
            use_values=False,
        )

    def test_attachments(self):
        a = LegacyAttachment(path="x")

        ids = []

        async def send_file(self, file_path, legacy_msg_id=None, *_a, **_k):
            ids.append(legacy_msg_id)
            m = Message()
            m.set_id(legacy_msg_id)
            return "", [m]

        orig = LegacyContact.send_file
        LegacyContact.send_file = send_file
        self.loop(self.juliet.send_files([], body="Hey"))
        assert not self.next_sent().get_id()

        self.loop(self.juliet.send_files([], body="Hey", legacy_msg_id="prout"))
        assert self.next_sent().get_id() == "prout"
        assert self.juliet

        self.loop(self.juliet.send_files([], body=""))
        assert self.next_sent() is None

        self.loop(self.juliet.send_files([a, a, a], body=""))
        assert ids.pop(-3) is None
        assert ids.pop(-2) is None
        assert ids.pop(-1) is None

        self.loop(self.juliet.send_files([a, a, a], legacy_msg_id="leg"))
        assert ids.pop(-3) == "leg"
        assert ids.pop(-2) is None
        assert ids.pop(-1) is None

        self.loop(self.juliet.send_files([], body="hoy"))
        assert not self.next_sent().get_id()

        self.loop(self.juliet.send_files([], body="hoy", legacy_msg_id="leg"))
        assert self.next_sent().get_id() == "leg"

        self.loop(
            self.juliet.send_files([], body="hoy", legacy_msg_id="leg", body_first=True)
        )
        assert self.next_sent().get_id() == "leg"

        self.loop(
            self.juliet.send_files(
                [a, a, a, a], body="hoy", legacy_msg_id="leg", body_first=True
            )
        )
        assert self.next_sent().get_id() == "leg"
        assert ids.pop(-4) is None
        assert ids.pop(-3) is None
        assert ids.pop(-2) is None
        assert ids.pop(-1) is None

        self.loop(
            self.juliet.send_files(
                [a, a, a, a], body="hoy", legacy_msg_id="leg", body_first=False
            )
        )
        assert ids.pop(-4) == "leg"
        assert ids.pop(-3) is None
        assert ids.pop(-2) is None
        assert ids.pop(-1) is None
        assert not self.next_sent().get_id()

        self.loop(self.juliet.send_files([a, a, a, a], body="hoy"))
        assert ids.pop(-4) is None
        assert ids.pop(-3) is None
        assert ids.pop(-2) is None
        assert ids.pop(-1) is None
        assert not self.next_sent().get_id()

        self.loop(self.juliet.send_files([a]))
        assert ids.pop(-1) is None

        self.loop(self.juliet.send_files([a], legacy_msg_id="leg"))
        assert ids.pop(-1) == "leg"

        LegacyContact.send_file = orig

    def test_gateway_message(self):
        session = self.get_romeo_session()
        session.send_gateway_message("Hello!")
        self.send(  # language=XML
            """
            <message xmlns="jabber:component:accept"
                     to="romeo@montague.lit"
                     from="aim.shakespeare.lit"
                     type="chat">
              <body>Hello!</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <store xmlns="urn:xmpp:hints" />
              <markable xmlns="urn:xmpp:chat-markers:0" />
            </message>
            """
        )


class TestPrivilege(SlidgeTest):
    plugin = globals()

    def setUp(self):
        super().setUp()
        with self.xmpp.store.session() as orm:
            user = GatewayUser(
                jid=JID("romeo@shakespeare.lit"),
                legacy_module_data={"username": "romeo", "city": ""},
                preferences={"sync_avatar": True, "sync_presence": True},
            )
            orm.add(user)
            orm.commit()

    def test_privilege(self):
        assert (
            self.xmpp["xep_0356"].granted_privileges["shakespeare.lit"] == Permissions()
        )
        self.recv(  # language=XML
            """
            <message to="aim.shakespeare.lit"
                     from="shakespeare.lit">
              <privilege xmlns="urn:xmpp:privilege:2">
                <perm access="roster"
                      type="both" />
                <perm access="message"
                      type="outgoing" />
              </privilege>
            </message>
            """
        )
        assert (
            self.xmpp["xep_0356"].granted_privileges["shakespeare.lit"].message
            == MessagePermission.OUTGOING
        )
        assert (
            self.xmpp["xep_0356"].granted_privileges["shakespeare.lit"].presence
            == PresencePermission.NONE
        )
        assert (
            self.xmpp["xep_0356"].granted_privileges["shakespeare.lit"].roster
            == RosterAccess.BOTH
        )

        session = BaseSession.get_self_or_unique_subclass().from_jid(
            JID("romeo@shakespeare.lit")
        )
        juliet = self.run_coro(
            session.contacts.by_jid(JID("juliet@aim.shakespeare.lit"))
        )
        juliet.send_text("body", 545, carbon=True)
        self.send(  # language=XML
            """
            <message to="shakespeare.lit"
                     from="aim.shakespeare.lit">
              <privilege xmlns="urn:xmpp:privilege:2">
                <forwarded xmlns="urn:xmpp:forward:0">
                  <message xmlns="jabber:client"
                           to="juliet@aim.shakespeare.lit"
                           type="chat"
                           from="romeo@shakespeare.lit"
                           id="545">
                    <body>body</body>
                    <store xmlns="urn:xmpp:hints" />
                    <markable xmlns="urn:xmpp:chat-markers:0" />
                    <active xmlns="http://jabber.org/protocol/chatstates" />
                  </message>
                </forwarded>
              </privilege>
            </message>
            """,
        )
        juliet.is_friend = True
        self.xmpp.loop.create_task(juliet.add_to_roster())
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="set"
                to="romeo@shakespeare.lit"
                from="aim.shakespeare.lit"
                id="1">
              <query xmlns="jabber:iq:roster">
                <item subscription="both"
                      jid="juliet@aim.shakespeare.lit">
                  <group>slidge</group>
                </item>
              </query>
            </iq>
            """
        )


class TestContact(SlidgeTest):
    plugin = globals()

    def setUp(self):
        super().setUp()
        with self.xmpp.store.session() as orm:
            user = GatewayUser(
                jid=JID("romeo@montague.lit/gajim").bare,
                legacy_module_data={"username": "romeo", "city": ""},
                preferences={"sync_avatar": True, "sync_presence": True},
            )
            orm.add(user)
            orm.execute(delete(Contact))
            orm.commit()
        romeo = self.get_romeo_session()
        romeo.logged = True
        romeo.contacts.ready.set_result(True)

    @staticmethod
    def get_romeo_session() -> Session:
        return BaseSession.get_self_or_unique_subclass().from_jid(
            JID("romeo@montague.lit")
        )

    @staticmethod
    def get_presence(ptype: str):
        return f"""
            <presence from="romeo@montague.lit"
                      to="juliet@aim.shakespeare.lit"
                      type="{ptype}" />
            """

    def get_contact(self, legacy_id: int):
        session = self.get_romeo_session()
        return self.run_coro(session.contacts.by_legacy_id(legacy_id))

    def get_juliet(self) -> LegacyContact:
        return self.get_contact(123)

    def get_new_friend(self) -> LegacyContact:
        return self.get_contact(456)

    def test_caps(self):
        juliet = self.get_juliet()
        juliet.is_friend = True
        juliet.online()
        self.send(  # language=XML
            f"""
            <presence xmlns="jabber:component:accept"
                      from="juliet@aim.shakespeare.lit/slidge"
                      to="romeo@montague.lit">
              <c xmlns="http://jabber.org/protocol/caps"
                 node="http://slixmpp.com/ver/{slix_version}"
                 hash="sha-1"
                 ver="UQAWii/9nNwlRpOcbTSGXm6r+aY=" />
              <priority>0</priority>
            </presence>
            """
        )

    def test_caps_extended(self):
        LegacyContact.REACTIONS_SINGLE_EMOJI = True
        LegacyContact.CORRECTION = False
        juliet = self.get_juliet()
        juliet.is_friend = True
        juliet.online()
        self.send(  # language=XML
            f"""
            <presence xmlns="jabber:component:accept"
                      from="juliet@aim.shakespeare.lit/slidge"
                      to="romeo@montague.lit">
              <c xmlns="http://jabber.org/protocol/caps"
                 node="http://slixmpp.com/ver/{slix_version}"
                 hash="sha-1"
                 ver="IFcuRXqYgRH7E9ZvZRxPhOrTH7E=" />
              <priority>0</priority>
            </presence>
            """
        )
        LegacyContact.REACTIONS_SINGLE_EMOJI = False
        LegacyContact.CORRECTION = True

    def test_vcard_temp(self):
        juliet = self.get_juliet()
        juliet.is_friend = True
        self.recv(  # language=XML
            """
            <iq from='romeo@montague.lit/orchard'
                to='juliet@aim.shakespeare.lit'
                type='get'>
              <vCard xmlns='vcard-temp' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                from="juliet@aim.shakespeare.lit"
                to="romeo@montague.lit/orchard"
                type="result"
                id="1">
              <vCard xmlns="vcard-temp" />
            </iq>
            """
        )
        juliet.set_vcard(full_name="Juliet Something")
        self.next_sent()
        self.recv(  # language=XML
            """
            <iq from='romeo@montague.lit/orchard'
                to='juliet@aim.shakespeare.lit'
                type='get'>
              <vCard xmlns='vcard-temp' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq from="juliet@aim.shakespeare.lit"
                to="romeo@montague.lit/orchard"
                type="result"
                id="2">
              <vCard xmlns="vcard-temp">
                <impp>
                  <uri xmlns="urn:ietf:params:xml:ns:vcard-4.0">xmpp:juliet@aim.shakespeare.lit</uri>
                </impp>
                <fn>
                  <text xmlns="urn:ietf:params:xml:ns:vcard-4.0">Juliet Something</text>
                </fn>
              </vCard>
            </iq>
            """,
            use_values=False,
        )

    def test_probe(self):
        juliet = self.get_juliet()
        probe = self.get_presence("probe")

        juliet.is_friend = True

        self.recv(probe)
        p = self.next_sent()
        assert p["type"] == "unavailable"

        juliet.online()
        assert self.next_sent()["type"] == "available"

        self.recv(probe)
        assert self.next_sent()["type"] == "available"

        juliet.is_friend = False
        self.recv(probe)
        p = self.next_sent()
        assert p["type"] == "unsubscribed"
        assert p["from"] == juliet.jid.bare
        assert self.next_sent() is None

    def test_user_subscribe_to_friend(self):
        juliet = self.get_juliet()
        juliet.is_friend = True
        sub = self.get_presence("subscribe")

        with unittest.mock.patch(
            "slidge.contact.LegacyContact.on_friend_request"
        ) as mock:
            self.recv(sub)
            mock.assert_not_awaited()
        p = self.next_sent()
        assert p["type"] == "subscribed"
        assert p["from"] == juliet.jid.bare
        assert self.next_sent() is None
        assert juliet.is_friend

    def test_user_subscribe_to_non_friend_accept(self):
        juliet = self.get_juliet()
        juliet.is_friend = False

        sub = self.get_presence("subscribe")

        with unittest.mock.patch(
            "slidge.contact.LegacyContact.on_friend_request"
        ) as mock:
            self.recv(sub)
            mock.assert_awaited_once()

        assert self.next_sent() is None
        assert not juliet.is_friend

        juliet.name = "JULIET"
        assert self.next_sent() is None
        self.run_coro(juliet.accept_friend_request())
        p = self.next_sent()
        assert p["type"] == "subscribed"
        assert p["from"] == juliet.jid.bare
        assert (
            self.next_sent()["pubsub_event"]["items"]["item"]["nick"]["nick"]
            == "JULIET"
        )
        assert self.next_sent() is None
        assert juliet.is_friend

    def test_user_subscribe_to_non_friend_reject(self):
        juliet = self.get_juliet()
        juliet.is_friend = False
        sub = self.get_presence("subscribe")

        with unittest.mock.patch(
            "slidge.contact.LegacyContact.on_friend_request"
        ) as mock:
            self.recv(sub)
            mock.assert_awaited_once()

        assert self.next_sent() is None

        juliet.name = "JULIET"
        assert self.next_sent() is None
        juliet.reject_friend_request()
        p = self.next_sent()
        assert p["from"] == juliet.jid.bare
        assert p["type"] == "unsubscribed"
        assert self.next_sent() is None
        assert not juliet.is_friend

    def test_juliet_send_friend_request_user_accepts(self):
        juliet = self.get_juliet()
        juliet.name = "JUJU"

        juliet.send_friend_request()
        p = self.next_sent()
        assert p["from"] == juliet.jid.bare
        assert p["type"] == "subscribe"
        assert p["to"] == "romeo@montague.lit"
        assert p["nick"]["nick"] == "JUJU"
        assert self.next_sent() is None

        with unittest.mock.patch(
            "slidge.contact.LegacyContact.on_friend_accept"
        ) as mock:
            self.recv(  # language=XML
                f"""
            <presence from='romeo@montague.lit/movim'
                      to='{juliet.jid.bare}'
                      type="subscribed" />
            """
            )
            mock.assert_awaited_once()
        assert self.next_sent() is None

    def test_juliet_send_friend_request_user_rejects(self):
        juliet = self.get_juliet()
        juliet.name = "JUJU"
        juliet.is_friend = False

        juliet.send_friend_request()
        p = self.next_sent()
        assert p["type"] == "subscribe"
        assert p["from"] == juliet.jid.bare
        assert p["to"] == "romeo@montague.lit"
        assert p["nick"]["nick"] == "JUJU"
        assert self.next_sent() is None

        with unittest.mock.patch(
            "slidge.contact.LegacyContact.on_friend_delete"
        ) as mock:
            self.recv(  # language=XML
                f"""
            <presence from='romeo@montague.lit/movim'
                      to='{juliet.jid.bare}'
                      type="unsubscribed" />
            """
            )
            mock.assert_not_awaited()

        assert self.next_sent() is None

        juliet.is_friend = True
        with unittest.mock.patch(
            "slidge.contact.LegacyContact.on_friend_delete"
        ) as mock:
            self.recv(  # language=XML
                f"""
            <presence from='romeo@montague.lit/movim'
                      to='{juliet.jid.bare}'
                      type="unsubscribed" />
            """
            )
            mock.assert_awaited_once()
        assert self.next_sent() is None

    def test_send_several_subscription_requests_in_a_row(self):
        juliet = self.get_juliet()
        juliet.online()
        juliet.send_friend_request()
        assert self.next_sent()["type"] == "subscribe"
        juliet.send_friend_request()
        assert self.next_sent()["type"] == "subscribe"

    def test_correct(self):
        juliet = self.get_juliet()

        juliet.correct("old_msg_id", "new content")
        msg = self.next_sent()
        assert msg["replace"]["id"] == "old_msg_id"
        assert msg["body"] == "new content"
        assert msg["id"] == ""

        juliet.correct("old_msg_id", "new content", correction_event_id="correction_id")
        msg = self.next_sent()
        assert msg["replace"]["id"] == "old_msg_id"
        assert msg["body"] == "new content"
        assert msg["id"] == "correction_id"

    def test_retract(self):
        juliet = self.get_juliet()

        juliet.retract("old_msg_id")
        # msg = self.next_sent()
        self.send(  # language=XML
            """
            <message type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     to="romeo@montague.lit">
              <body>/me retracted the message old_msg_id</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <store xmlns="urn:xmpp:hints" />
              <fallback xmlns="urn:xmpp:fallback:0"
                        for="urn:xmpp:message-retract:1" />
              <retract xmlns="urn:xmpp:message-retract:1"
                       id="old_msg_id" />
              <replace id='old_msg_id'
                       xmlns='urn:xmpp:message-correct:0' />
            </message>
            """
        )

    def test_presence(self):
        juliet = self.get_juliet()
        juliet.is_friend = True

        now = datetime.datetime.now(tz=datetime.timezone.utc)
        a_week_ago = now - datetime.timedelta(days=7)
        juliet.away(status="Bye bye", last_seen=now)

        p = self.next_sent()
        assert p["status"] == f"Bye bye -- Last seen {now:%A %H:%M %p GMT}"
        assert p["idle"]["since"] == now
        assert p["show"] == "away"

        juliet.extended_away(status="Bye bye", last_seen=a_week_ago)
        p = self.next_sent()
        assert p["status"] == f"Bye bye -- Last seen {a_week_ago:%b %-d %Y}"
        assert p["idle"]["since"] == a_week_ago
        assert p["show"] == "xa"

        juliet.busy(last_seen=now)
        p = self.next_sent()
        assert p["status"] == f"Last seen {now:%A %H:%M %p GMT}"
        assert p["idle"]["since"] == now
        assert p["show"] == "dnd"

        juliet.busy(last_seen=a_week_ago)
        p = self.next_sent()
        assert p["status"] == f"Last seen {a_week_ago:%b %-d %Y}"
        assert p["idle"]["since"] == a_week_ago

    def test_disco_bare(self):
        juliet = self.get_juliet()
        juliet.is_friend = True
        self.recv(  # language=XML
            f"""
            <iq type='get'
                from='romeo@montague.lit'
                to="juliet@aim.shakespeare.lit">
              <query xmlns='http://jabber.org/protocol/disco#info' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq type="result"
                from="juliet@aim.shakespeare.lit"
                to="romeo@montague.lit"
                id="1">
              <query xmlns="http://jabber.org/protocol/disco#info">
                <identity category="pubsub"
                          type="pep"
                          name="SLIDGE TEST" />
                <identity category="account"
                          type="registered"
                          name="SLIDGE TEST" />
                <feature var="http://jabber.org/protocol/pubsub" />
                <feature var="http://jabber.org/protocol/pubsub#retrieve-items" />
                <feature var="http://jabber.org/protocol/pubsub#subscribe" />
              </query>
            </iq>
            """
        )

    def test_disco_resource(self):
        juliet = self.get_juliet()
        juliet.is_friend = True
        self.recv(  # language=XML
            f"""
            <iq type='get'
                from='romeo@montague.lit'
                to="juliet@aim.shakespeare.lit/slidge">
              <query xmlns='http://jabber.org/protocol/disco#info' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq type="result"
                from="juliet@aim.shakespeare.lit/slidge"
                to="romeo@montague.lit"
                id="1">
              <query xmlns="http://jabber.org/protocol/disco#info">
                <identity category="client"
                          type="pc"
                          name="SLIDGE TEST" />
                <feature var="http://jabber.org/protocol/chatstates" />
                <feature var="urn:xmpp:receipts" />
                <feature var="urn:xmpp:message-correct:0" />
                <feature var="urn:xmpp:chat-markers:0" />
                <feature var="jabber:x:oob" />
                <feature var="urn:xmpp:reactions:0" />
                <feature var="urn:xmpp:message-retract:0" />
                <feature var="urn:xmpp:reply:0" />
                <feature var="urn:ietf:params:xml:ns:vcard-4.0" />
                <feature var="jabber:x:conference" />
              </query>
            </iq>
            """
        )

    def test_body_needing_escaping(self):
        self.get_juliet().send_text("\x1b[0;31mRED")
        self.send(  # language=XML
            """
            <message type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     to="romeo@montague.lit">
              <body>�[0;31mRED</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <store xmlns="urn:xmpp:hints" />
              <markable xmlns="urn:xmpp:chat-markers:0" />
            </message>
            """
        )

class TestCarbon(SlidgeTest):
    plugin = globals()

    def setUp(self):
        super().setUp()
        with self.xmpp.store.session() as orm:
            user = GatewayUser(
                jid=JID("romeo@shakespeare.lit").bare,
                legacy_module_data={"username": "romeo", "city": ""},
                preferences={"sync_avatar": True, "sync_presence": True},
            )
            orm.add(user)
            orm.commit()
        self.recv(  # language=XML
            """
            <message to="aim.shakespeare.lit"
                     from="shakespeare.lit">
              <privilege xmlns="urn:xmpp:privilege:2">
                <perm access="roster"
                      type="both" />
                <perm access="message"
                      type="outgoing" />
              </privilege>
            </message>
            """
        )
        self.get_romeo_session().logged = True

    @staticmethod
    def get_romeo_session() -> Session:
        return BaseSession.get_self_or_unique_subclass().from_jid(
            JID("romeo@shakespeare.lit")
        )

    def get_juliet(self) -> LegacyContact:
        session = self.get_romeo_session()
        return self.run_coro(session.contacts.by_legacy_id(123))

    def test_carbon_send_file(self):
        orig = AttachmentMixin._AttachmentMixin__get_url

        async def get_url(self, att, *a, **k):
            return False, att.path, "URL"

        AttachmentMixin._AttachmentMixin__get_url = get_url

        juliet = self.get_juliet()
        juliet.send_text("TEXT", 445, carbon=True)
        self.send(  # language=XML
            """
            <message xmlns="jabber:component:accept"
                     to="shakespeare.lit"
                     from="aim.shakespeare.lit"
                     type="normal">
              <privilege xmlns="urn:xmpp:privilege:2">
                <forwarded xmlns="urn:xmpp:forward:0">
                  <message xmlns="jabber:client"
                           type="chat"
                           from="romeo@shakespeare.lit"
                           to="juliet@aim.shakespeare.lit"
                           id='445'>
                    <body>TEXT</body>
                    <active xmlns="http://jabber.org/protocol/chatstates" />
                    <store xmlns="urn:xmpp:hints" />
                    <markable xmlns="urn:xmpp:chat-markers:0" />
                  </message>
                </forwarded>
              </privilege>
            </message>
            """
        )
        with tempfile.NamedTemporaryFile("w+") as f:
            f.write("test")
            f.seek(0)
            self.run_coro(
                juliet.send_file(Path(f.name), legacy_msg_id=446, carbon=True)
            )
            stamp = format_datetime(
                datetime.datetime.fromtimestamp(Path(f.name).stat().st_mtime)
            )
            self.send(  # language=XML
                f"""
            <message xmlns="jabber:component:accept"
                     to="shakespeare.lit"
                     from="aim.shakespeare.lit"
                     type="normal">
              <privilege xmlns="urn:xmpp:privilege:2">
                <forwarded xmlns="urn:xmpp:forward:0">
                  <message xmlns="jabber:client"
                           type="chat"
                           from="romeo@shakespeare.lit"
                           to="juliet@aim.shakespeare.lit"
                           id="446">
                    <reference xmlns="urn:xmpp:reference:0"
                               type="data">
                      <media-sharing xmlns="urn:xmpp:sims:1">
                        <sources>
                          <reference xmlns="urn:xmpp:reference:0"
                                     uri="URL"
                                     type="data" />
                        </sources>
                        <file xmlns="urn:xmpp:jingle:apps:file-transfer:5">
                          <name>{Path(f.name).name}</name>
                          <size>4</size>
                          <date>{stamp}</date>
                          <hash xmlns="urn:xmpp:hashes:2"
                                algo="sha-256">n4bQgYhMfWWaL+qgxVrQFaO/TxsrC4Is0V1sFbDwCgg=</hash>
                        </file>
                      </media-sharing>
                    </reference>
                    <fallback xmlns="urn:xmpp:fallback:0"
                              for="urn:xmpp:sfs:0">
                      <body />
                    </fallback>
                    <file-sharing xmlns="urn:xmpp:sfs:0">
                      <sources>
                        <url-data xmlns="http://jabber.org/protocol/url-data"
                                  target="URL" />
                      </sources>
                      <file xmlns="urn:xmpp:file:metadata:0">
                        <name>{Path(f.name).name}</name>
                        <size>4</size>
                        <date>{stamp}</date>
                        <hash xmlns="urn:xmpp:hashes:2"
                              algo="sha-256">n4bQgYhMfWWaL+qgxVrQFaO/TxsrC4Is0V1sFbDwCgg=</hash>
                      </file>
                    </file-sharing>
                    <x xmlns="jabber:x:oob">
                      <url>URL</url>
                    </x>
                    <body>URL</body>
                  </message>
                </forwarded>
              </privilege>
            </message>
            """
            )
        AttachmentMixin._AttachmentMixin__get_url = orig

    def test_carbon_ignore(self):
        """
        Ensure we don't react to stanzas we send to ourselves.
        """
        juliet = self.get_juliet()
        juliet.send_text("TEXT", 123456, carbon=True)
        self.send(  # language=XML
            """
            <message xmlns="jabber:component:accept"
                     to="shakespeare.lit"
                     from="aim.shakespeare.lit"
                     type="normal">
              <privilege xmlns="urn:xmpp:privilege:2">
                <forwarded xmlns="urn:xmpp:forward:0">
                  <message xmlns="jabber:client"
                           type="chat"
                           from="romeo@shakespeare.lit"
                           to="juliet@aim.shakespeare.lit"
                           id='123456'>
                    <body>TEXT</body>
                    <active xmlns="http://jabber.org/protocol/chatstates" />
                    <store xmlns="urn:xmpp:hints" />
                    <markable xmlns="urn:xmpp:chat-markers:0" />
                  </message>
                </forwarded>
              </privilege>
            </message>
            """
        )
        with unittest.mock.patch("test_shakespeare.Session.on_text") as mock:
            mock.get_id = lambda *a: "msg_id"
            self.recv(  # language=XML
                """
            <message type="chat"
                     from="romeo@shakespeare.lit"
                     to="juliet@aim.shakespeare.lit"
                     id='123456'>
              <body>TEXT</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <store xmlns="urn:xmpp:hints" />
              <markable xmlns="urn:xmpp:chat-markers:0" />
            </message>
            """
            )
            mock.assert_not_awaited()


class TestUserGetsOnline(Base):
    def setUp(self):
        super().setUp()
        self.get_romeo_session().contacts.ready.set_result(True)
        self.juliet.is_friend = True
        self.juliet.added_to_roster = True
        self.juliet.name = "Juliet"
        self.send(  # language=XML
            """
            <message type="headline"
                     from="juliet@aim.shakespeare.lit"
                     to="romeo@montague.lit">
              <event xmlns="http://jabber.org/protocol/pubsub#event">
                <items node="http://jabber.org/protocol/nick">
                  <item>
                    <nick xmlns="http://jabber.org/protocol/nick">Juliet</nick>
                  </item>
                </items>
              </event>
            </message>
            """,
            use_values=False,
        )

    def _assert_send_nick(self):
        self.send(  # language=XML
            """
            <message type="headline"
                     from="juliet@aim.shakespeare.lit"
                     to="romeo@montague.lit/cheogram">
              <event xmlns="http://jabber.org/protocol/pubsub#event">
                <items node="http://jabber.org/protocol/nick">
                  <item>
                    <nick xmlns="http://jabber.org/protocol/nick">Juliet</nick>
                  </item>
                </items>
              </event>
            </message>
            """,
            use_values=False,
        )

    def test_user_online_without_caps(self):
        self.recv(  # language=XML
            """
            <presence from="romeo@montague.lit/cheogram"
                      to="juliet@aim.shakespeare.lit" />
            """
        )
        self.send(  # language=XML
            """
            <iq id="1"
                from="aim.shakespeare.lit"
                to="romeo@montague.lit/cheogram"
                type="get">
              <query xmlns="http://jabber.org/protocol/disco#info" />
            </iq>
            """
        )
        self.recv(  # language=XML
            """
            <iq id="1"
                type="result"
                to="aim.shakespeare.lit"
                from="romeo@montague.lit/cheogram">
              <query xmlns='http://jabber.org/protocol/disco#info'>
                <feature var='http://jabber.org/protocol/nick+notify' />
              </query>
            </iq>
            """
        )
        self._assert_send_nick()
        assert self.next_sent() is None

    def test_user_online_with_caps(self):
        with unittest.mock.patch(
            "slixmpp.plugins.xep_0115.caps.XEP_0115.get_caps"
        ) as get_caps:
            get_caps.return_value = DiscoInfo(
                xml=ET.fromstring(  # language=XML
                    """
            <query xmlns='http://jabber.org/protocol/disco#info'>
              <feature var='http://jabber.org/protocol/nick+notify' />
            </query>
            """
                )
            )
            self.recv(  # language=XML
                """
            <presence from="romeo@montague.lit/cheogram"
                      to="juliet@aim.shakespeare.lit">
              <c xmlns='http://jabber.org/protocol/caps'
                 hash='sha-1'
                 node='test'
                 ver='VERSTRING' />
            </presence>
            """
            )
        self._assert_send_nick()
        assert self.next_sent() is None

    def test_user_online_with_caps_no_nick_notify(self):
        with unittest.mock.patch(
            "slixmpp.plugins.xep_0115.caps.XEP_0115.get_caps"
        ) as get_caps:
            get_caps.return_value = DiscoInfo(
                xml=ET.fromstring(  # language=XML
                    """
            <query xmlns='http://jabber.org/protocol/disco#info'>
              <feature var='http://jabber.org/protocol/whatevs+notify' />
            </query>
            """
                )
            )
            self.recv(  # language=XML
                """
            <presence from="romeo@montague.lit/cheogram"
                      to="juliet@aim.shakespeare.lit">
              <c xmlns='http://jabber.org/protocol/caps'
                 hash='sha-1'
                 node='test'
                 ver='VERSTRING' />
            </presence>
            """
            )
        assert self.next_sent() is None


class TestUserPresence(Base):
    def setUp(self):
        self.patcher = unittest.mock.patch(
            "slidge.core.session.BaseSession.on_presence"
        )
        self.presence = self.patcher.start()
        super().setUp()

    def tearDown(self):
        super().tearDown()
        self.patcher.stop()

    def test_user_presence(self):
        self.recv(  # language=XML
            """
            <presence from="romeo@montague.lit/cheogram"
                      to="aim.shakespeare.lit" />
            """
        )
        self.presence.assert_called_with(
            "cheogram",
            "available",
            "",
            {"cheogram": {"status": "", "show": "", "priority": 0}},
            {"status": "", "show": "", "priority": 0},
        )
        self.presence.assert_awaited_once()
        self.presence.reset_mock()
        self.recv(  # language=XML
            """
            <presence from="romeo@montague.lit/gajim"
                      to="aim.shakespeare.lit">
              <show>away</show>
              <status>I use gajim, yay!</status>
            </presence>
            """
        )
        self.presence.assert_called_with(
            "gajim",
            "away",
            "I use gajim, yay!",
            {
                "cheogram": {"status": "", "show": "", "priority": 0},
                "gajim": {"status": "I use gajim, yay!", "show": "away", "priority": 0},
            },
            {"status": "I use gajim, yay!", "show": "", "priority": 0},
        )
        self.recv(  # language=XML
            """
            <presence from="romeo@montague.lit/gajim"
                      to="aim.shakespeare.lit"
                      type="unavailable" />
            """
        )
        self.presence.assert_called_with(
            "gajim",
            "unavailable",
            "",
            {
                "cheogram": {"status": "", "show": "", "priority": 0},
            },
            {"status": "", "show": "", "priority": 0},
        )


log = logging.getLogger(__name__)



from typing import Any

from slixmpp import JID, Message
from slixmpp.xmlstream import StanzaBase

from ....contact import LegacyContact
from ....group.room import LegacyMUC
from ....util.types import Recipient
from ..util import DispatcherMixin, exceptions_to_xmpp_errors, get_recipient


class MarkerMixin(DispatcherMixin):
    __slots__: list[str] = []

    def __init__(self, xmpp) -> None:
        super().__init__(xmpp)
        xmpp.add_event_handler("marker_displayed", self.on_marker_displayed)
        xmpp.add_event_handler(
            "message_displayed_synchronization_publish",
            self.on_message_displayed_synchronization_publish,
        )

    @exceptions_to_xmpp_errors
    async def on_marker_displayed(self, msg: StanzaBase) -> None:
        assert isinstance(msg, Message)
        session = await self._get_session(msg)

        e: Recipient = await get_recipient(session, msg)
        legacy_thread = await self._xmpp_to_legacy_thread(session, msg, e)
        to_mark = self.__to_mark(e, msg["displayed"]["id"])
        for xmpp_id in to_mark:
            await session.on_displayed(
                e, self._xmpp_msg_id_to_legacy(session, xmpp_id, e), legacy_thread
            )
        if isinstance(e, LegacyMUC):
            await e.echo(msg, None)

    def __to_mark(
        self, chat: LegacyContact[Any] | LegacyMUC[Any, Any, Any, Any], msg_id: str
    ) -> list[str]:
        if self.xmpp.MARK_ALL_MESSAGES:
            return chat.pop_unread_xmpp_ids_up_to(msg_id)
        else:
            return [msg_id]

    @exceptions_to_xmpp_errors
    async def on_message_displayed_synchronization_publish(
        self, msg: StanzaBase
    ) -> None:
        assert isinstance(msg, Message)
        chat_jid = JID(msg["pubsub_event"]["items"]["item"]["id"])
        if chat_jid.server != self.xmpp.boundjid.bare:
            return

        session = await self._get_session(msg, timeout=None)

        if chat_jid == self.xmpp.boundjid.bare:
            return

        chat = await session.get_contact_or_group_or_participant(chat_jid)
        if not isinstance(chat, LegacyMUC):
            session.log.debug("Ignoring non-groupchat MDS event")
            return

        stanza_id = msg["pubsub_event"]["items"]["item"]["displayed"]["stanza_id"]["id"]
        to_mark = self.__to_mark(chat, stanza_id)
        for xmpp_id in to_mark:
            await session.on_displayed(
                chat, self._xmpp_msg_id_to_legacy(session, xmpp_id, chat)
            )



import pytest

from conftest import AvatarFixtureMixin
from slixmpp import JID, Iq

from slidge import BaseGateway, BaseSession, GatewayUser, LegacyRoster, LegacyBookmarks
from slidge.core.session import _sessions
from slidge.util.test import SlidgeTest
from slidge.util.types import LegacyUserIdType


class SomeType:
    def __init__(self, a: int, b: int):
        self.a = a
        self.b = b

    @classmethod
    def from_str(cls, s: str):
        a, b = (int(x) for x in s.split("-"))
        return SomeType(a, b)

    def __str__(self):
        return f"{self.a}-{self.b}"

class Gateway(BaseGateway):
    COMPONENT_NAME = "A test"
    LEGACY_CONTACT_ID_TYPE = SomeType.from_str
    LEGACY_ROOM_ID_TYPE = SomeType.from_str
    GROUPS = True


class Session(BaseSession):
    async def login(self):
        return "YUP"


class Roster(LegacyRoster):
    async def legacy_id_to_jid_username(self, legacy_id: LegacyUserIdType) -> str:
        return f"{legacy_id.a}-{legacy_id.b}"

    async def jid_username_to_legacy_id(self, jid_username: str) -> LegacyUserIdType:
        return SomeType.from_str(jid_username)


class Bookmarks(LegacyBookmarks):
    async def legacy_id_to_jid_username(self, legacy_id: LegacyUserIdType) -> str:
        return f"{legacy_id.a}-{legacy_id.b}"

    async def jid_username_to_legacy_id(self, jid_username: str) -> LegacyUserIdType:
        return SomeType.from_str(jid_username)

    async def fill(self):
        pass

@pytest.mark.usefixtures("avatar")
class TestLegacyTypeConversion(AvatarFixtureMixin, SlidgeTest):
    plugin = globals()
    xmpp: Gateway

    def setUp(self):
        super().setUp()
        with self.xmpp.store.session() as orm:
            user = GatewayUser(
                jid=JID("romeo@montague.lit/gajim").bare,
                legacy_module_data={"username": "romeo", "city": ""},
                preferences={"sync_avatar": True, "sync_presence": True},
            )
            orm.add(user)
            orm.commit()
        self.run_coro(
            self.xmpp._BaseGateway__dispatcher._on_user_register(
                Iq(sfrom="romeo@montague.lit/gajim")
            )
        )
        welcome = self.next_sent()
        assert welcome["body"]
        stanza = self.next_sent()
        assert "logging in" in stanza["status"].lower(), stanza
        stanza = self.next_sent()
        assert "syncing contacts" in stanza["status"].lower(), stanza
        stanza = self.next_sent()
        assert "syncing groups" in stanza["status"].lower(), stanza
        probe = self.next_sent()
        assert probe.get_type() == "probe"
        stanza = self.next_sent()
        assert "yup" in stanza["status"].lower(), stanza

        self.send(  # language=XML
            """
            <iq type="get"
                to="romeo@montague.lit"
                id="1"
                from="aim.shakespeare.lit">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <items node="urn:xmpp:avatar:metadata" />
              </pubsub>
            </iq>
            """
        )

    def tearDown(self):
        super().tearDown()
        _sessions.clear()

    @property
    def romeo_session(self) -> Session:
        return BaseSession.get_self_or_unique_subclass().from_jid(
            JID("romeo@montague.lit")
        )

    def test_contact(self):
        contact = self.run_coro(self.romeo_session.contacts.by_legacy_id(SomeType(1, 2)))
        assert contact.legacy_id.a == 1
        assert contact.legacy_id.b == 2

    def test_muc(self):
        muc = self.run_coro(self.romeo_session.bookmarks.by_legacy_id(SomeType(1, 2)))
        assert muc.legacy_id.a == 1
        assert muc.legacy_id.b == 2



import logging
import typing
from contextlib import contextmanager

import sqlalchemy as sa
from sqlalchemy.exc import IntegrityError

from ...db.models import Base, Contact, Participant, Room

if typing.TYPE_CHECKING:
    from slidge import BaseGateway


class DBMixin:
    stored: Base
    xmpp: "BaseGateway"
    log: logging.Logger

    def merge(self) -> None:
        with self.xmpp.store.session() as orm:
            self.stored = orm.merge(self.stored)

    def commit(self, merge: bool = False) -> None:
        with self.xmpp.store.session(expire_on_commit=False) as orm:
            if merge:
                self.log.debug("Merging %s", self.stored)
                self.stored = orm.merge(self.stored)
                self.log.debug("Merged %s", self.stored)
            orm.add(self.stored)
            self.log.debug("Committing to DB")
            orm.commit()


class UpdateInfoMixin(DBMixin):
    """
    This mixin just adds a context manager that prevents commiting to the DB
    on every attribute change.
    """

    stored: Contact | Room
    xmpp: "BaseGateway"
    log: logging.Logger

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._updating_info = False
        self.__deserialize()

    def __deserialize(self):
        if self.stored.extra_attributes is not None:
            self.deserialize_extra_attributes(self.stored.extra_attributes)

    def refresh(self) -> None:
        with self.xmpp.store.session(expire_on_commit=False) as orm:
            orm.add(self.stored)
            orm.refresh(self.stored)
            self.__deserialize()

    def serialize_extra_attributes(self) -> dict | None:
        """
        If you want custom attributes of your instance to be stored persistently
        to the DB, here is where you have to return them as a dict to be used in
        `deserialize_extra_attributes()`.

        """
        return None

    def deserialize_extra_attributes(self, data: dict) -> None:
        """
        This is where you get the dict that you passed in
        `serialize_extra_attributes()`.

        ⚠ Since it is serialized as json, dictionary keys are converted to strings!
        Be sure to convert to other types if necessary.
        """
        pass

    @contextmanager
    def updating_info(self, merge=True):
        self._updating_info = True
        yield
        self._updating_info = False
        self.stored.updated = True
        try:
            self.commit(merge=merge)
        except IntegrityError as e:
            from slidge.group import LegacyMUC

            if not isinstance(self, LegacyMUC):
                raise e
            if not self._ALL_INFO_FILLED_ON_STARTUP:
                raise e
            with self.xmpp.store.session(expire_on_commit=False) as orm:
                if self.stored.id is None:
                    self.stored.id = self.xmpp.store.rooms.get(
                        orm, self.user_pk, legacy_id=str(self.legacy_id)
                    ).id
                merged = orm.merge(self.stored)
                resources = set()
                participants: list[Participant] = []
                for participant in merged.participants:
                    if participant.resource in resources:
                        self.session.log.warning("ditching: %s", participant)
                        continue
                    resources.add(participant.resource)
                merged.participants = participants
                self.stored = merged
                orm.add(self.stored)
                orm.commit()

    def commit(self, merge: bool = False) -> None:
        if self._updating_info:
            self.log.debug("Not updating %s right now", self.stored)
        else:
            self.stored.extra_attributes = self.serialize_extra_attributes()
            super().commit(merge=merge)

    def update_stored_attribute(self, **kwargs) -> None:
        for key, value in kwargs.items():
            setattr(self.stored, key, value)
        if self._updating_info:
            return
        with self.xmpp.store.session() as orm:
            orm.execute(
                sa.update(self.stored.__class__)
                .where(self.stored.__class__.id == self.stored.id)
                .values(**kwargs)
            )
            orm.commit()



import hashlib
from asyncio import Task
from pathlib import Path
from typing import TYPE_CHECKING, Optional

from PIL import UnidentifiedImageError
from slixmpp import JID
from sqlalchemy.orm.exc import DetachedInstanceError

from ...db.avatar import CachedAvatar, avatar_cache
from ...db.models import Contact, Room
from ...util.types import AnyBaseSession, Avatar
from .db import UpdateInfoMixin

if TYPE_CHECKING:
    from ..pubsub import PepAvatar


class AvatarMixin(UpdateInfoMixin):
    """
    Mixin for XMPP entities that have avatars that represent them.

    Both :py:class:`slidge.LegacyContact` and :py:class:`slidge.LegacyMUC` use
    :py:class:`.AvatarMixin`.
    """

    jid: JID = NotImplemented
    session: AnyBaseSession = NotImplemented
    stored: Contact | Room

    def __init__(self) -> None:
        super().__init__()
        self._set_avatar_task: Task | None = None

    @property
    def avatar(self) -> Avatar | None:
        """
        This property can be used to set or unset the avatar.

        Unlike the awaitable :method:`.set_avatar`, it schedules the update for
        later execution and is not blocking
        """
        try:
            if self.stored.avatar is None:
                return None
        except DetachedInstanceError:
            self.merge()
            if self.stored.avatar is None:
                return None
        if self.stored.avatar.legacy_id is None:
            unique_id = None
        else:
            unique_id = self.session.xmpp.AVATAR_ID_TYPE(self.stored.avatar.legacy_id)
        return Avatar(
            unique_id=unique_id,
            url=self.stored.avatar.url,
        )

    @avatar.setter
    def avatar(self, avatar: Avatar | Path | str | None) -> None:
        avatar = convert_avatar(avatar)
        if self._set_avatar_task:
            self._set_avatar_task.cancel()
        self.session.log.debug("Setting avatar with property")
        self._set_avatar_task = self.session.create_task(self.set_avatar(avatar))

    async def __has_changed(self, avatar: Avatar | None) -> bool:
        if self.avatar is None:
            return avatar is not None
        if avatar is None:
            return self.avatar is not None

        if self.avatar.unique_id is not None and avatar.unique_id is not None:
            return self.avatar.unique_id != avatar.unique_id

        if (
            self.avatar.url is not None
            and avatar.url is not None
            and self.avatar.url == avatar.url
        ):
            return await avatar_cache.url_modified(avatar.url)

        if avatar.path is not None:
            cached = self.get_cached_avatar()
            if cached is not None:
                return cached.path.read_bytes() != avatar.path.read_bytes()

        return True

    async def set_avatar(
        self, avatar: Avatar | Path | str | None = None, delete: bool = False
    ) -> None:
        """
        Set an avatar for this entity

        :param avatar: The avatar. Should ideally come with a legacy network-wide unique
            ID
        :param delete: If the avatar is provided as a Path, whether to delete
            it once used or not.
        """
        avatar = convert_avatar(avatar)

        if avatar is not None and avatar.unique_id is None and avatar.data is not None:
            self.session.log.debug("Hashing bytes to generate a unique ID")
            avatar = Avatar(
                data=avatar.data, unique_id=hashlib.sha512(avatar.data).hexdigest()
            )

        if not await self.__has_changed(avatar):
            return

        if avatar is None:
            cached_avatar = None
        else:
            try:
                cached_avatar = await avatar_cache.convert_or_get(avatar)
            except UnidentifiedImageError:
                self.session.log.warning("%s is not a valid image", avatar)
                cached_avatar = None
            except Exception as e:
                self.session.log.error("Failed to set avatar %s: %s", avatar, e)
                cached_avatar = None

        if delete:
            if avatar is None or avatar.path is None:
                self.session.log.warning(
                    "Requested avatar path delete, but no path provided"
                )
            else:
                avatar.path.unlink()

        stored_avatar = None if cached_avatar is None else cached_avatar.stored
        if not self._updating_info:
            with self.xmpp.store.session() as orm:
                with orm.no_autoflush:
                    self.stored = orm.merge(self.stored)
                    orm.refresh(self.stored)

        self.stored.avatar = stored_avatar
        self.commit(merge=True)

        self._post_avatar_update(cached_avatar)

    def get_cached_avatar(self) -> Optional["CachedAvatar"]:
        try:
            if self.stored.avatar is None:
                return None
        except DetachedInstanceError:
            self.merge()
            if self.stored.avatar is None:
                return None
        return avatar_cache.get(self.stored.avatar)

    def get_avatar(self) -> Optional["PepAvatar"]:
        cached_avatar = self.get_cached_avatar()
        if cached_avatar is None:
            return None
        from ..pubsub import PepAvatar

        item = PepAvatar()
        item.set_avatar_from_cache(cached_avatar)
        return item

    def _post_avatar_update(self, cached_avatar: Optional["CachedAvatar"]) -> None:
        raise NotImplementedError


def convert_avatar(
    avatar: Avatar | Path | str | None, unique_id: str | None = None
) -> Avatar | None:
    if isinstance(avatar, Path):
        return Avatar(path=avatar, unique_id=unique_id)
    if isinstance(avatar, str):
        return Avatar(url=avatar)
    if avatar is None or all(x is None for x in avatar):
        return None
    return avatar



import re
from asyncio import Task, sleep
from datetime import UTC, datetime, timedelta
from functools import partial
from typing import TYPE_CHECKING

from slixmpp.types import PresenceShows, PresenceTypes
from sqlalchemy.orm.exc import DetachedInstanceError

from ...db.models import Contact, Participant
from ...util.types import CachedPresence
from .base import BaseSender
from .db import DBMixin

if TYPE_CHECKING:
    from ..session import BaseSession


class _NoChange(Exception):
    pass


_FRIEND_REQUEST_PRESENCES = {"subscribe", "unsubscribe", "subscribed", "unsubscribed"}
_UPDATE_LAST_SEEN_FALLBACK_TASKS = dict[int, Task]()
_ONE_WEEK_SECONDS = 3600 * 24 * 7


async def _update_last_seen_fallback(session: "BaseSession", contact_pk: int) -> None:
    await sleep(_ONE_WEEK_SECONDS)
    with session.xmpp.store.session() as orm:
        stored = orm.get(Contact, contact_pk)
        if stored is None:
            return
        contact = session.contacts.from_store(stored)
    contact.send_last_presence(force=True, no_cache_online=False)


def _clear_last_seen_task(contact_pk: int, _task) -> None:
    try:
        del _UPDATE_LAST_SEEN_FALLBACK_TASKS[contact_pk]
    except KeyError:
        pass


class PresenceMixin(BaseSender, DBMixin):
    _ONLY_SEND_PRESENCE_CHANGES = False

    # this attribute actually only exists for contacts and not participants
    _updating_info: bool
    stored: Contact | Participant

    def __init__(self, *a, **k) -> None:
        super().__init__(*a, **k)
        # this is only used when a presence is set during Contact.update_info(),
        # when the contact does not have a DB primary key yet, and is written
        # to DB at the end of update_info()
        self.cached_presence: CachedPresence | None = None

    def __is_contact(self) -> bool:
        return isinstance(self.stored, Contact)

    def __stored(self) -> Contact | None:
        if self.__is_contact():
            assert isinstance(self.stored, Contact)
            return self.stored
        else:
            assert isinstance(self.stored, Participant)
            try:
                return self.stored.contact
            except DetachedInstanceError:
                with self.xmpp.store.session() as orm:
                    orm.add(self.stored)
                    if self.stored.contact is None:
                        return None
                    orm.refresh(self.stored.contact)
                    orm.merge(self.stored)
                    return self.stored.contact

    @property
    def __contact_pk(self) -> int | None:
        stored = self.__stored()
        return None if stored is None else stored.id

    def _get_last_presence(self) -> CachedPresence | None:
        stored = self.__stored()
        if stored is None or not stored.cached_presence:
            return None
        return CachedPresence(
            None if stored.last_seen is None else stored.last_seen.replace(tzinfo=UTC),
            stored.ptype,  # type:ignore
            stored.pstatus,
            stored.pshow,  # type:ignore
        )

    def _store_last_presence(self, new: CachedPresence) -> None:
        if self.__is_contact():
            contact = self
        elif (contact := getattr(self, "contact", None)) is None:  # type:ignore[assignment]
            return
        contact.update_stored_attribute(  # type:ignore[attr-defined]
            cached_presence=True,
            **new._asdict(),
        )

    def _make_presence(
        self,
        *,
        last_seen: datetime | None = None,
        force: bool = False,
        bare: bool = False,
        ptype: PresenceTypes | None = None,
        pstatus: str | None = None,
        pshow: PresenceShows | None = None,
    ):
        if last_seen and last_seen.tzinfo is None:
            last_seen = last_seen.astimezone(UTC)

        old = self._get_last_presence()

        if ptype not in _FRIEND_REQUEST_PRESENCES:
            new = CachedPresence(
                last_seen=last_seen, ptype=ptype, pstatus=pstatus, pshow=pshow
            )
            if old != new:
                if hasattr(self, "muc") and ptype == "unavailable":
                    stored = self.__stored()
                    if stored is not None:
                        stored.cached_presence = False
                        self.commit(merge=True)
                else:
                    self._store_last_presence(new)
            if old and not force and self._ONLY_SEND_PRESENCE_CHANGES:
                if old == new:
                    self.session.log.debug("Presence is the same as cached")
                    raise _NoChange
                self.session.log.debug(
                    "Presence is not the same as cached: %s vs %s", old, new
                )

        p = self.xmpp.make_presence(
            pfrom=self.jid.bare if bare else self.jid,
            ptype=ptype,
            pshow=pshow,
            pstatus=pstatus,
        )
        if last_seen:
            # it's ugly to check for the presence of this string, but a better fix is more work
            if not re.match(
                ".*Last seen .*", p["status"]
            ) and self.session.user.preferences.get("last_seen_fallback", True):
                last_seen_fallback, recent = get_last_seen_fallback(last_seen)
                if p["status"]:
                    p["status"] = p["status"] + " -- " + last_seen_fallback
                else:
                    p["status"] = last_seen_fallback
                pk = self.__contact_pk
                if recent and pk is not None:
                    # if less than a week, we use sth like 'Last seen: Monday, 8:05",
                    # but if lasts more than a week, this is not very informative, so
                    # we need to force resend an updated presence status
                    task = _UPDATE_LAST_SEEN_FALLBACK_TASKS.get(pk)
                    if task is not None:
                        task.cancel()
                    task = self.session.create_task(
                        _update_last_seen_fallback(self.session, pk)
                    )
                    _UPDATE_LAST_SEEN_FALLBACK_TASKS[pk] = task
                    task.add_done_callback(partial(_clear_last_seen_task, pk))
            p["idle"]["since"] = last_seen
        return p

    def send_last_presence(
        self, force: bool = False, no_cache_online: bool = False
    ) -> None:
        if (cache := self._get_last_presence()) is None:
            if force:
                if no_cache_online:
                    self.online()
                else:
                    self.offline()
            return
        self._send(
            self._make_presence(
                last_seen=cache.last_seen,
                force=True,
                ptype=cache.ptype,
                pshow=cache.pshow,
                pstatus=cache.pstatus,
            )
        )

    def online(
        self,
        status: str | None = None,
        last_seen: datetime | None = None,
    ) -> None:
        """
        Send an "online" presence from this contact to the user.

        :param status: Arbitrary text, details of the status, eg: "Listening to Britney Spears"
        :param last_seen: For :xep:`0319`
        """
        try:
            self._send(self._make_presence(pstatus=status, last_seen=last_seen))
        except _NoChange:
            pass

    def away(
        self,
        status: str | None = None,
        last_seen: datetime | None = None,
    ) -> None:
        """
        Send an "away" presence from this contact to the user.

        This is a global status, as opposed to :meth:`.LegacyContact.inactive`
        which concerns a specific conversation, ie a specific "chat window"

        :param status: Arbitrary text, details of the status, eg: "Gone to fight capitalism"
        :param last_seen: For :xep:`0319`
        """
        try:
            self._send(
                self._make_presence(pstatus=status, pshow="away", last_seen=last_seen)
            )
        except _NoChange:
            pass

    def extended_away(
        self,
        status: str | None = None,
        last_seen: datetime | None = None,
    ) -> None:
        """
        Send an "extended away" presence from this contact to the user.

        This is a global status, as opposed to :meth:`.LegacyContact.inactive`
        which concerns a specific conversation, ie a specific "chat window"

        :param status: Arbitrary text, details of the status, eg: "Gone to fight capitalism"
        :param last_seen: For :xep:`0319`
        """
        try:
            self._send(
                self._make_presence(pstatus=status, pshow="xa", last_seen=last_seen)
            )
        except _NoChange:
            pass

    def busy(
        self,
        status: str | None = None,
        last_seen: datetime | None = None,
    ) -> None:
        """
        Send a "busy" (ie, "dnd") presence from this contact to the user,

        :param status: eg: "Trying to make sense of XEP-0100"
        :param last_seen: For :xep:`0319`
        """
        try:
            self._send(
                self._make_presence(pstatus=status, pshow="dnd", last_seen=last_seen)
            )
        except _NoChange:
            pass

    def offline(
        self,
        status: str | None = None,
        last_seen: datetime | None = None,
    ) -> None:
        """
        Send an "offline" presence from this contact to the user.

        :param status: eg: "Trying to make sense of XEP-0100"
        :param last_seen: For :xep:`0319`
        """
        try:
            self._send(
                self._make_presence(
                    pstatus=status, ptype="unavailable", last_seen=last_seen
                )
            )
        except _NoChange:
            pass


def get_last_seen_fallback(last_seen: datetime) -> tuple[str, bool]:
    now = datetime.now(tz=UTC)
    if now - last_seen < timedelta(days=7):
        return f"Last seen {last_seen:%A %H:%M %p GMT}", True
    else:
        return f"Last seen {last_seen:%b %-d %Y}", False



import logging
import uuid
import warnings
from collections.abc import Collection
from copy import copy
from datetime import UTC, datetime
from typing import TYPE_CHECKING, Optional

from slixmpp import Iq, Message

from ..db.models import ArchivedMessage, ArchivedMessageSource, Room
from ..db.store import SlidgeStore
from ..util.archive_msg import HistoryMessage
from ..util.types import HoleBound

if TYPE_CHECKING:
    from .participant import LegacyParticipant


class MessageArchive:
    def __init__(self, room: Room, store: SlidgeStore) -> None:
        self.room = room
        self.__rooms_store = store.rooms
        self.__store = store.mam

    def add(
        self,
        msg: Message,
        participant: Optional["LegacyParticipant"] = None,
        archive_only: bool = False,
        legacy_msg_id=None,
    ) -> None:
        """
        Add a message to the archive if it is deemed archivable

        :param msg:
        :param participant:
        :param archive_only:
        :param legacy_msg_id:
        """
        if not archivable(msg):
            return
        new_msg = copy(msg)
        if participant and not participant.muc.is_anonymous:
            new_msg["muc"]["role"] = participant.role or "participant"
            new_msg["muc"]["affiliation"] = participant.affiliation or "member"
            if participant.contact:
                new_msg["muc"]["jid"] = participant.contact.jid.bare
            elif participant.is_user:
                new_msg["muc"]["jid"] = participant.user_jid.bare
            elif participant.is_system:
                new_msg["muc"]["jid"] = participant.muc.jid
            else:
                warnings.warn(
                    f"No real JID for participant '{participant.nickname}' in '{self.room.name}'"
                )
                new_msg["muc"]["jid"] = (
                    f"{uuid.uuid4()}@{participant.xmpp.boundjid.bare}"
                )

        if "replace" in msg:
            legacy_msg_id = None

        with self.__store.session(expire_on_commit=False) as orm:
            if self.room.id is None:
                self.room = self.__rooms_store.get(
                    orm, self.room.user_account_id, self.room.legacy_id
                )
            self.__store.add_message(
                orm,
                self.room.id,
                HistoryMessage(new_msg),
                archive_only,
                None if legacy_msg_id is None else str(legacy_msg_id),
            )
            orm.commit()

    def __iter__(self):
        return iter(self.get_all())

    @staticmethod
    def __to_bound(stored: ArchivedMessage):
        return HoleBound(
            stored.legacy_id,  # type:ignore
            stored.timestamp.replace(tzinfo=UTC),
        )

    def get_hole_bounds(self) -> tuple[HoleBound | None, HoleBound | None]:
        with self.__store.session() as orm:
            most_recent = self.__store.get_most_recent_with_legacy_id(orm, self.room.id)
            if most_recent is None:
                return None, None
            if most_recent.source == ArchivedMessageSource.BACKFILL:
                # most recent = only backfill, fetch everything since last backfill
                return self.__to_bound(most_recent), None

            most_recent_back_filled = self.__store.get_most_recent_with_legacy_id(
                orm, self.room.id, ArchivedMessageSource.BACKFILL
            )
            if most_recent_back_filled is None:
                # group was never back-filled, fetch everything before first live
                least_recent_live = self.__store.get_first(orm, self.room.id, True)
                assert least_recent_live is not None
                return None, self.__to_bound(least_recent_live)

            assert most_recent_back_filled.legacy_id is not None
            least_recent_live = self.__store.get_least_recent_with_legacy_id_after(
                orm, self.room.id, most_recent_back_filled.legacy_id
            )
            assert least_recent_live is not None
            # this is a hole caused by slidge downtime
            return self.__to_bound(most_recent_back_filled), self.__to_bound(
                least_recent_live
            )

    def get_all(
        self,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
        before_id: str | None = None,
        after_id: str | None = None,
        ids: Collection[str] = (),
        last_page_n: int | None = None,
        sender: str | None = None,
        flip: bool = False,
    ):
        with self.__store.session() as orm:
            yield from self.__store.get_messages(
                orm,
                self.room.id,
                before_id=before_id,
                after_id=after_id,
                ids=ids,
                last_page_n=last_page_n,
                sender=sender,
                start_date=start_date,
                end_date=end_date,
                flip=flip,
            )

    async def send_metadata(self, iq: Iq) -> None:
        """
        Send archive extent, as per the spec

        :param iq:
        :return:
        """
        reply = iq.reply()
        with self.__store.session() as orm:
            messages = self.__store.get_first_and_last(orm, self.room.id)
        if messages:
            for x, m in [("start", messages[0]), ("end", messages[-1])]:
                reply["mam_metadata"][x]["id"] = m.id
                reply["mam_metadata"][x]["timestamp"] = m.sent_on.replace(tzinfo=UTC)
        else:
            reply.enable("mam_metadata")
        reply.send()


def archivable(msg: Message) -> bool:
    """
    Determine if a message stanza is worth archiving, ie, convey meaningful
    info

    :param msg:
    :return:
    """

    if msg.get_plugin("no-store", check=True):
        return False

    if msg.get_plugin("no-permanent-store", check=True):
        return False

    if msg.get_plugin("store", check=True):
        return True

    if msg["body"]:
        return True

    if msg.get_plugin("retract", check=True):
        return True

    if msg.get_plugin("reactions", check=True):
        return True

    if msg.get_plugin("displayed", check=True):
        return True

    if msg["thread"] and msg["subject"]:
        return True

    return False


log = logging.getLogger(__name__)



"""
This module handles the registration :term:`Command`, which is a necessary
step for a JID to become a slidge :term:`User`.
"""

import asyncio
import functools
import tempfile
from enum import IntEnum
from typing import Any

import qrcode
from slixmpp import JID, Iq
from slixmpp.exceptions import XMPPError

from ..core import config
from ..db import GatewayUser
from ..util.types import UserPreferences
from .base import Command, CommandAccess, Form, FormField, FormValues
from .user import Preferences


class RegistrationType(IntEnum):
    """
    An :class:`Enum` to define the registration flow.
    """

    SINGLE_STEP_FORM = 0
    """
    1 step, 1 form, the only flow compatible with :xep:`0077`.
    Using this, the whole flow is defined
    by :attr:`slidge.BaseGateway.REGISTRATION_FIELDS` and
    :attr:`.REGISTRATION_INSTRUCTIONS`.
    """

    QRCODE = 10
    """
    The registration requires flashing a QR code in an official client.
    See :meth:`slidge.BaseGateway.send_qr`, :meth:`.get_qr_text`
    and :meth:`.confirm_qr`.
    """

    TWO_FACTOR_CODE = 20
    """
    The registration requires confirming login with a 2FA code,
    eg something received by email or SMS to finalize the authentication.
    See :meth:`.validate_two_factor_code`.
    """


class TwoFactorNotRequired(Exception):
    """
    Should be raised in :meth:`slidge.BaseGateway.validate` if the code is not
    required after all. This can happen for a :term:`Legacy Network` where 2FA
    is optional.
    """

    pass


class Register(Command):
    NAME = "📝 Register to the gateway"
    HELP = "Link your JID to this gateway"
    NODE = "jabber:iq:register"
    CHAT_COMMAND = "register"
    ACCESS = CommandAccess.NON_USER

    SUCCESS_MESSAGE = "Success, welcome!"

    def _finalize(
        self, form_values: UserPreferences, _session, ifrom: JID, user: GatewayUser, *_
    ) -> str:
        user.preferences = form_values  # type: ignore
        self.xmpp.store.users.update(user)
        self.xmpp.event("user_register", Iq(sfrom=ifrom.bare))
        return self.SUCCESS_MESSAGE

    async def run(self, _session, ifrom: JID, *_):
        self.xmpp.raise_if_not_allowed_jid(ifrom)
        return Form(
            title=f"Registration to '{self.xmpp.COMPONENT_NAME}'",
            instructions=self.xmpp.REGISTRATION_INSTRUCTIONS,
            fields=self.xmpp.REGISTRATION_FIELDS,
            handler=self.register,
        )

    async def register(self, form_values: dict[str, Any], _session, ifrom: JID):
        two_fa_needed = True
        try:
            data = await self.xmpp.user_prevalidate(ifrom, form_values)
        except ValueError as e:
            raise XMPPError("bad-request", str(e))
        except TwoFactorNotRequired:
            data = None
            if self.xmpp.REGISTRATION_TYPE == RegistrationType.TWO_FACTOR_CODE:
                two_fa_needed = False
            else:
                raise

        user = GatewayUser(
            jid=JID(ifrom.bare),
            legacy_module_data=form_values if data is None else data,
        )

        if self.xmpp.REGISTRATION_TYPE == RegistrationType.SINGLE_STEP_FORM or (
            self.xmpp.REGISTRATION_TYPE == RegistrationType.TWO_FACTOR_CODE
            and not two_fa_needed
        ):
            return await self.preferences(user)

        if self.xmpp.REGISTRATION_TYPE == RegistrationType.TWO_FACTOR_CODE:
            return Form(
                title=self.xmpp.REGISTRATION_2FA_TITLE,
                instructions=self.xmpp.REGISTRATION_2FA_INSTRUCTIONS,
                fields=[FormField("code", label="Code", required=True)],
                handler=functools.partial(self.two_fa, user=user),
            )

        elif self.xmpp.REGISTRATION_TYPE == RegistrationType.QRCODE:
            self.xmpp.qr_pending_registrations[  # type:ignore
                user.jid.bare
            ] = self.xmpp.loop.create_future()
            qr_text = await self.xmpp.get_qr_text(user)
            qr = qrcode.make(qr_text)
            with tempfile.NamedTemporaryFile(
                suffix=".png", delete=config.NO_UPLOAD_METHOD != "move"
            ) as f:
                qr.save(f.name)
                img_url, _ = await self.xmpp.send_file(f.name, mto=ifrom)
            if img_url is None:
                raise XMPPError(
                    "internal-server-error", "Slidge cannot send attachments"
                )
            self.xmpp.send_text(qr_text, mto=ifrom)
            return Form(
                title="Flash this",
                instructions="Flash this QR in the appropriate place",
                fields=[
                    FormField(
                        "qr_img",
                        type="fixed",
                        value=qr_text,
                        image_url=img_url,
                    ),
                    FormField(
                        "qr_text",
                        type="fixed",
                        value=qr_text,
                        label="Text encoded in the QR code",
                    ),
                    FormField(
                        "qr_img_url",
                        type="fixed",
                        value=img_url,
                        label="URL of the QR code image",
                    ),
                ],
                handler=functools.partial(self.qr, user=user),
            )

    async def two_fa(
        self, form_values: FormValues, _session, _ifrom, user: GatewayUser
    ):
        assert isinstance(form_values["code"], str)
        data = await self.xmpp.validate_two_factor_code(user, form_values["code"])
        if data is not None:
            user.legacy_module_data.update(data)
        return await self.preferences(user)

    async def qr(self, _form_values: FormValues, _session, _ifrom, user: GatewayUser):
        try:
            data = await asyncio.wait_for(
                self.xmpp.qr_pending_registrations[user.jid.bare],  # type:ignore
                config.QR_TIMEOUT,
            )
        except TimeoutError:
            raise XMPPError(
                "remote-server-timeout",
                (
                    "It does not seem that the QR code was correctly used, "
                    "or you took too much time"
                ),
            )
        if data is not None:
            user.legacy_module_data.update(data)
        return await self.preferences(user)

    async def preferences(self, user: GatewayUser) -> Form:
        return Form(
            title="Preferences",
            instructions=Preferences.HELP,
            fields=self.xmpp.PREFERENCES,
            handler=functools.partial(self._finalize, user=user),  # type:ignore
            timeout_handler=functools.partial(self._preferences_timeout, user=user),  # type:ignore
        )

    def _preferences_timeout(self, user: GatewayUser) -> None:
        self.xmpp.event("user_register", Iq(sfrom=user.jid.bare))
        self.xmpp.store.users.update(user)
        self.xmpp.send_message(
            mfrom=self.xmpp.boundjid.bare,
            mto=user.jid,
            mbody="You did not choose your preferences in time, falling back to defaults. "
            "You can change preferences later with the 'preferences' command.\n"
            + self.SUCCESS_MESSAGE,
        )



"""
This module extends slixmpp.ComponentXMPP to make writing new LegacyClients easier
"""

import abc
import asyncio
import logging
import re
import tempfile
from collections.abc import Callable, Sequence
from copy import copy
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import aiohttp
import qrcode
from slixmpp import JID, ComponentXMPP, Iq, Message, Presence
from slixmpp.exceptions import IqError, IqTimeout, XMPPError
from slixmpp.plugins.xep_0004 import Form
from slixmpp.plugins.xep_0060.stanza import OwnerAffiliation
from slixmpp.plugins.xep_0356.permissions import IqPermission
from slixmpp.plugins.xep_0356.privilege import PrivilegedIqError
from slixmpp.types import MessageTypes
from slixmpp.xmlstream.xmlstream import NotConnectedError

from slidge import LegacyContact, command  # noqa: F401
from slidge.command.adhoc import AdhocProvider
from slidge.command.admin import Exec
from slidge.command.base import Command, FormField
from slidge.command.chat_command import ChatCommandProvider
from slidge.command.register import RegistrationType
from slidge.core import config
from slidge.core.dispatcher.session_dispatcher import SessionDispatcher
from slidge.core.mixins import MessageMixin
from slidge.core.mixins.attachment import AttachmentMixin
from slidge.core.mixins.avatar import convert_avatar
from slidge.core.pubsub import PubSubComponent
from slidge.core.session import BaseSession
from slidge.db import GatewayUser, SlidgeStore
from slidge.db.avatar import CachedAvatar, avatar_cache
from slidge.db.meta import JSONSerializable
from slidge.slixfix.delivery_receipt import DeliveryReceipt
from slidge.slixfix.roster import RosterBackend
from slidge.util import SubclassableOnce
from slidge.util.types import Avatar, MessageOrPresenceTypeVar
from slidge.util.util import timeit

if TYPE_CHECKING:
    pass


class BaseGateway(
    ComponentXMPP,
    MessageMixin,
    SubclassableOnce,
    abc.ABC,
):
    """
    The gateway component, handling registrations and un-registrations.

    On slidge launch, a singleton is instantiated, and it will be made available
    to public classes such :class:`.LegacyContact` or :class:`.BaseSession` as the
    ``.xmpp`` attribute.

    Must be subclassed by a legacy module to set up various aspects of the XMPP
    component behaviour, such as its display name or welcome message, via
    class attributes :attr:`.COMPONENT_NAME` :attr:`.WELCOME_MESSAGE`.

    Abstract methods related to the registration process must be overriden
    for a functional :term:`Legacy Module`:

    - :meth:`.validate`
    - :meth:`.validate_two_factor_code`
    - :meth:`.get_qr_text`
    - :meth:`.confirm_qr`

    NB: Not all of these must be overridden, it depends on the
    :attr:`REGISTRATION_TYPE`.

    The other methods, such as :meth:`.send_text` or :meth:`.react` are the same
    as those of :class:`.LegacyContact` and :class:`.LegacyParticipant`, because
    the component itself is also a "messaging actor", ie, an :term:`XMPP Entity`.
    For these methods, you need to specify the JID of the recipient with the
    `mto` parameter.

    Since it inherits from :class:`slixmpp.componentxmpp.ComponentXMPP`,you also
    have a hand on low-level XMPP interactions via slixmpp methods, e.g.:

    .. code-block:: python

        self.send_presence(
            pfrom="somebody@component.example.com",
            pto="someonwelse@anotherexample.com",
        )

    However, you should not need to do so often since the classes of the plugin
    API provides higher level abstractions around most commonly needed use-cases, such
    as sending messages, or displaying a custom status.

    """

    COMPONENT_NAME: str = NotImplemented
    """Name of the component, as seen in service discovery by XMPP clients"""
    COMPONENT_TYPE: str = ""
    """Type of the gateway, should follow https://xmpp.org/registrar/disco-categories.html"""
    COMPONENT_AVATAR: Avatar | Path | str | None = None
    """
    Path, bytes or URL used by the component as an avatar.
    """

    REGISTRATION_FIELDS: Sequence[FormField] = [
        FormField(var="username", label="User name", required=True),
        FormField(var="password", label="Password", required=True, private=True),
    ]
    """
    Iterable of fields presented to the gateway user when registering using :xep:`0077`
    `extended <https://xmpp.org/extensions/xep-0077.html#extensibility>`_ by :xep:`0004`.
    """
    REGISTRATION_INSTRUCTIONS: str = "Enter your credentials"
    """
    The text presented to a user that wants to register (or modify) their legacy account
    configuration.
    """
    REGISTRATION_TYPE: RegistrationType = RegistrationType.SINGLE_STEP_FORM
    """
    This attribute determines how users register to the gateway, ie, how they
    login to the :term:`legacy service <Legacy Service>`.
    The credentials are then stored persistently, so this process should happen
    once per user (unless they unregister).

    The registration process always start with a basic data form (:xep:`0004`)
    presented to the user.
    But the legacy login flow might require something more sophisticated, see
    :class:`.RegistrationType` for more details.
    """

    REGISTRATION_2FA_TITLE = "Enter your 2FA code"
    REGISTRATION_2FA_INSTRUCTIONS = (
        "You should have received something via email or SMS, or something"
    )
    REGISTRATION_QR_INSTRUCTIONS = "Flash this code or follow this link"

    PREFERENCES = [
        FormField(
            var="sync_presence",
            label="Propagate your XMPP presence to the legacy network.",
            value="true",
            required=True,
            type="boolean",
        ),
        FormField(
            var="sync_avatar",
            label="Propagate your XMPP avatar to the legacy network.",
            value="true",
            required=True,
            type="boolean",
        ),
        FormField(
            var="always_invite_when_adding_bookmarks",
            label="Send an invitation to join MUCs after adding them to the bookmarks.",
            value="true",
            required=True,
            type="boolean",
        ),
        FormField(
            var="last_seen_fallback",
            label="Use contact presence status message to show when they were last seen.",
            value="true",
            required=True,
            type="boolean",
        ),
        FormField(
            var="roster_push",
            label="Add contacts to your roster.",
            value="true",
            required=True,
            type="boolean",
        ),
        FormField(
            var="reaction_fallback",
            label="Receive fallback messages for reactions (for legacy XMPP clients)",
            value="false",
            required=True,
            type="boolean",
        ),
    ]

    ROSTER_GROUP: str = "slidge"
    """
    Name of the group assigned to a :class:`.LegacyContact` automagically
    added to the :term:`User`'s roster with :meth:`.LegacyContact.add_to_roster`.
    """
    WELCOME_MESSAGE = (
        "Thank you for registering. Type 'help' to list the available commands, "
        "or just start messaging away!"
    )
    """
    A welcome message displayed to users on registration.
    This is useful notably for clients that don't consider component JIDs as a
    valid recipient in their UI, yet still open a functional chat window on
    incoming messages from components.
    """

    SEARCH_FIELDS: Sequence[FormField] = [
        FormField(var="first", label="First name", required=True),
        FormField(var="last", label="Last name", required=True),
        FormField(var="phone", label="Phone number", required=False),
    ]
    """
    Fields used for searching items via the component, through :xep:`0055` (jabber search).
    A common use case is to allow users to search for legacy contacts by something else than
    their usernames, eg their phone number.

    Plugins should implement search by overriding :meth:`.BaseSession.search`
    (restricted to registered users).

    If there is only one field, it can also be used via the ``jabber:iq:gateway`` protocol
    described in :xep:`0100`. Limitation: this only works if the search request returns
    one result item, and if this item has a 'jid' var.
    """
    SEARCH_TITLE: str = "Search for legacy contacts"
    """
    Title of the search form.
    """
    SEARCH_INSTRUCTIONS: str = ""
    """
    Instructions of the search form.
    """

    MARK_ALL_MESSAGES = False
    """
    Set this to True for :term:`legacy networks <Legacy Network>` that expects
    read marks for *all* messages and not just the latest one that was read
    (as most XMPP clients will only send a read mark for the latest msg).
    """

    PROPER_RECEIPTS = False
    """
    Set this to True if the legacy service provides a real equivalent of message delivery receipts
    (:xep:`0184`), meaning that there is an event thrown when the actual device of a contact receives
    a message. Make sure to call Contact.received() adequately if this is set to True.
    """

    GROUPS = False

    mtype: MessageTypes = "chat"
    is_group = False
    _can_send_carbon = False
    store: SlidgeStore

    AVATAR_ID_TYPE: Callable[[str], Any] = str
    """
    Modify this if the legacy network uses unique avatar IDs that are not strings.

    This is required because we store those IDs as TEXT in the persistent SQL DB.
    The callable specified here will receive is responsible for converting the
    serialised-as-text version of the avatar unique ID back to the proper type.
    Common example: ``int``.
    """
    # FIXME: do we really need this since we have session.xmpp_to_legacy_msg_id?
    #        (maybe we do)
    LEGACY_MSG_ID_TYPE: Callable[[str], Any] = str
    """
    Modify this if the legacy network uses unique message IDs that are not strings.

    This is required because we store those IDs as TEXT in the persistent SQL DB.
    The callable specified here will receive is responsible for converting the
    serialised-as-text version of the message unique ID back to the proper type.
    Common example: ``int``.
    """
    LEGACY_CONTACT_ID_TYPE: Callable[[str], Any] = str
    """
    Modify this if the legacy network uses unique contact IDs that are not strings.

    This is required because we store those IDs as TEXT in the persistent SQL DB.
    The callable specified here is responsible for converting the
    serialised-as-text version of the contact unique ID back to the proper type.
    Common example: ``int``.
    """
    LEGACY_ROOM_ID_TYPE: Callable[[str], Any] = str
    """
    Modify this if the legacy network uses unique room IDs that are not strings.

    This is required because we store those IDs as TEXT in the persistent SQL DB.
    The callable specified here is responsible for converting the
    serialised-as-text version of the room unique ID back to the proper type.
    Common example: ``int``.
    """

    DB_POOL_SIZE: int = 5
    """
    Size of the queue pool for sqlalchemy engine. Typically, when using python async
    libraries, this does not need to be changed.
    Change that if your gateway use separate threads to call into slidge. The value of
    this parameter should be equal or greater than the potential number of threads.
    """

    http: aiohttp.ClientSession
    avatar: CachedAvatar | None = None

    def __init__(self) -> None:
        if config.COMPONENT_NAME:
            self.COMPONENT_NAME = config.COMPONENT_NAME
        if config.WELCOME_MESSAGE:
            self.WELCOME_MESSAGE = config.WELCOME_MESSAGE
        self.log = log
        self.datetime_started = datetime.now()
        self.xmpp = self  # ugly hack to work with the BaseSender mixin :/
        self.default_ns = "jabber:component:accept"
        super().__init__(
            config.JID,
            config.SECRET,
            config.SERVER,
            config.PORT,
            plugin_whitelist=SLIXMPP_PLUGINS,
            plugin_config={
                "xep_0077": {
                    "form_fields": None,
                    "form_instructions": self.REGISTRATION_INSTRUCTIONS,
                    "enable_subscription": self.REGISTRATION_TYPE
                    == RegistrationType.SINGLE_STEP_FORM,
                },
                "xep_0100": {
                    "component_name": self.COMPONENT_NAME,
                    "type": self.COMPONENT_TYPE,
                },
                "xep_0184": {
                    "auto_ack": False,
                    "auto_request": False,
                },
                "xep_0363": {
                    "upload_service": config.UPLOAD_SERVICE,
                },
            },
            fix_error_ns=True,
        )
        self.loop.set_exception_handler(self.__exception_handler)
        self.loop.create_task(self.__set_http())
        self.has_crashed: bool = False
        self.use_origin_id = False

        if config.USER_JID_VALIDATOR is None:
            config.USER_JID_VALIDATOR = f".*@{self.infer_real_domain()}"
            log.info(
                "No USER_JID_VALIDATOR was set, using '%s'.",
                config.USER_JID_VALIDATOR,
            )
        self.jid_validator: re.Pattern = re.compile(config.USER_JID_VALIDATOR)
        self.qr_pending_registrations = dict[str, asyncio.Future[dict | None]]()

        self.register_plugins()
        self.__setup_legacy_module_subclasses()

        self.get_session_from_stanza: Callable[
            [Message | Presence | Iq], BaseSession
        ] = self._session_cls.from_stanza
        self.get_session_from_user: Callable[[GatewayUser], BaseSession] = (
            self._session_cls.from_user
        )

        self.__register_slixmpp_events()
        self.__register_slixmpp_api()
        self.roster.set_backend(RosterBackend(self))

        self.register_plugin("pubsub", {"component_name": self.COMPONENT_NAME})
        self.pubsub: PubSubComponent = self["pubsub"]
        self.delivery_receipt: DeliveryReceipt = DeliveryReceipt(self)

        # with this we receive user avatar updates
        self.plugin["xep_0030"].add_feature("urn:xmpp:avatar:metadata+notify")

        self.plugin["xep_0030"].add_feature("urn:xmpp:chat-markers:0")

        if self.GROUPS:
            self.plugin["xep_0030"].add_feature("http://jabber.org/protocol/muc")
            self.plugin["xep_0030"].add_feature("urn:xmpp:mam:2")
            self.plugin["xep_0030"].add_feature("urn:xmpp:mam:2#extended")
            self.plugin["xep_0030"].add_feature(self.plugin["xep_0421"].namespace)
            self.plugin["xep_0030"].add_feature(self["xep_0317"].stanza.NS)
            self.plugin["xep_0030"].add_identity(
                category="conference",
                name=self.COMPONENT_NAME,
                itype="text",
                jid=self.boundjid,
            )

        # why does mypy need these type annotations? no idea
        self.__adhoc_handler: AdhocProvider = AdhocProvider(self)
        self.__chat_commands_handler: ChatCommandProvider = ChatCommandProvider(self)

        self.__dispatcher = SessionDispatcher(self)

        self.__register_commands()

        MessageMixin.__init__(self)  # ComponentXMPP does not call super().__init__()

    def __setup_legacy_module_subclasses(self):
        from ..contact.roster import LegacyRoster
        from ..group.bookmarks import LegacyBookmarks
        from ..group.room import LegacyMUC, LegacyParticipant

        session_cls: type[BaseSession] = cast(
            type[BaseSession], BaseSession.get_unique_subclass()
        )
        contact_cls = LegacyContact.get_self_or_unique_subclass()
        muc_cls = LegacyMUC.get_self_or_unique_subclass()
        participant_cls = LegacyParticipant.get_self_or_unique_subclass()
        bookmarks_cls = LegacyBookmarks.get_self_or_unique_subclass()
        roster_cls = LegacyRoster.get_self_or_unique_subclass()

        if contact_cls.REACTIONS_SINGLE_EMOJI:  # type:ignore[attr-defined]
            form = Form()
            form["type"] = "result"
            form.add_field(
                "FORM_TYPE", "hidden", value="urn:xmpp:reactions:0:restrictions"
            )
            form.add_field("max_reactions_per_user", value="1", type="number")
            form.add_field("scope", value="domain")
            self.plugin["xep_0128"].add_extended_info(data=form)

        session_cls.xmpp = self  # type:ignore[attr-defined]
        contact_cls.xmpp = self  # type:ignore[attr-defined]
        muc_cls.xmpp = self  # type:ignore[attr-defined]

        self._session_cls = session_cls
        session_cls._bookmarks_cls = bookmarks_cls  # type:ignore[misc,assignment]
        session_cls._roster_cls = roster_cls  # type:ignore[misc,assignment]
        LegacyRoster._contact_cls = contact_cls  # type:ignore[misc]
        LegacyBookmarks._muc_cls = muc_cls  # type:ignore[misc]
        LegacyMUC._participant_cls = participant_cls  # type:ignore[misc]

    async def kill_session(self, jid: JID) -> None:
        await self._session_cls.kill_by_jid(jid)

    async def __set_http(self) -> None:
        self.http = aiohttp.ClientSession()
        if getattr(self, "_test_mode", False):
            return
        avatar_cache.http = self.http

    def __register_commands(self) -> None:
        for cls in Command.subclasses:
            if any(x is NotImplemented for x in [cls.CHAT_COMMAND, cls.NODE, cls.NAME]):
                log.debug("Not adding command '%s' because it looks abstract", cls)
                continue
            if cls is Exec:
                if config.DEV_MODE:
                    log.warning(r"/!\ DEV MODE ENABLED /!\\")
                else:
                    continue
            c = cls(self)
            log.debug("Registering %s", cls)
            self.__adhoc_handler.register(c)
            self.__chat_commands_handler.register(c)

    def __exception_handler(self, loop: asyncio.AbstractEventLoop, context) -> None:
        """
        Called when a task created by loop.create_task() raises an Exception

        :param loop:
        :param context:
        :return:
        """
        log.debug("Context in the exception handler: %s", context)
        exc = context.get("exception")
        if exc is None:
            log.debug("No exception in this context: %s", context)
        elif isinstance(exc, SystemExit):
            log.debug("SystemExit called in an asyncio task")
        else:
            log.error("Crash in an asyncio task: %s", context)
            log.exception("Crash in task", exc_info=exc)
            self.has_crashed = True
            loop.stop()

    def __register_slixmpp_events(self) -> None:
        self.del_event_handler("presence_subscribe", self._handle_subscribe)
        self.del_event_handler("presence_unsubscribe", self._handle_unsubscribe)
        self.del_event_handler("presence_subscribed", self._handle_subscribed)
        self.del_event_handler("presence_unsubscribed", self._handle_unsubscribed)
        self.del_event_handler(
            "roster_subscription_request", self._handle_new_subscription
        )
        self.del_event_handler("presence_probe", self._handle_probe)
        self.add_event_handler("session_start", self.__on_session_start)
        self.add_event_handler("privileges_advertised", self.__on_privileges_advertised)
        self.__upload_service_found = asyncio.Event()

    def __register_slixmpp_api(self) -> None:
        def with_session(func, commit: bool = True):
            def wrapped(*a, **kw):
                with self.store.session() as orm:
                    res = func(orm, *a, **kw)
                    if commit:
                        orm.commit()
                    return res

            return wrapped

        self.plugin["xep_0231"].api.register(
            with_session(self.store.bob.get_bob, False), "get_bob"
        )
        self.plugin["xep_0231"].api.register(
            with_session(self.store.bob.set_bob), "set_bob"
        )
        self.plugin["xep_0231"].api.register(
            with_session(self.store.bob.del_bob), "del_bob"
        )

    @property  # type: ignore
    def jid(self):  # type:ignore[override]
        # Override to avoid slixmpp deprecation warnings.
        return self.boundjid

    async def __on_session_start(self, event) -> None:
        log.debug("Gateway session start: %s", event)

        await self.__setup_attachments()

        # prevents XMPP clients from considering the gateway as an HTTP upload
        disco = self.plugin["xep_0030"]
        await disco.del_feature(feature="urn:xmpp:http:upload:0", jid=self.boundjid)
        await self.plugin["xep_0115"].update_caps(jid=self.boundjid)

        if self.COMPONENT_AVATAR is not None:
            log.debug("Setting gateway avatar…")
            avatar = convert_avatar(self.COMPONENT_AVATAR, "!!---slidge---special---")
            assert avatar is not None
            try:
                cached_avatar = await avatar_cache.convert_or_get(avatar)
            except Exception as e:
                log.exception("Could not set the component avatar.", exc_info=e)
                cached_avatar = None
            else:
                assert cached_avatar is not None
                self.avatar = cached_avatar
        else:
            cached_avatar = None

        with self.store.session() as orm:
            for user in orm.query(GatewayUser).all():
                # TODO: before this, we should check if the user has removed us from their roster
                #       while we were offline and trigger unregister from there. Presence probe does not seem
                #       to work in this case, there must be another way. privileged entity could be used
                #       as last resort.
                try:
                    await self["xep_0100"].add_component_to_roster(user.jid)
                    await self.__add_component_to_mds_whitelist(user.jid)
                except (IqError, IqTimeout) as e:
                    # TODO: remove the user when this happens? or at least
                    # this can happen when the user has unsubscribed from the XMPP server
                    log.warning(
                        "Error with user %s, not logging them automatically",
                        user,
                        exc_info=e,
                    )
                    continue
                session = self._session_cls.from_user(user)
                session.create_task(self.login_wrap(session))
                if cached_avatar is not None:
                    await self.pubsub.broadcast_avatar(
                        self.boundjid.bare, session.user_jid, cached_avatar
                    )

        log.info("Slidge has successfully started")

    async def __on_privileges_advertised(self, _) -> None:
        await self.__upload_service_found.wait()
        if config.UPLOAD_SERVICE and self.xmpp["xep_0356"].granted_privileges[
            config.SERVER
        ].iq.get(self.plugin["xep_0363"].stanza.Request.namespace) in (
            IqPermission.SET,
            IqPermission.BOTH,
            IqPermission.GET,
        ):
            AttachmentMixin.PRIVILEGED_UPLOAD = True

    async def __setup_attachments(self) -> None:
        if config.NO_UPLOAD_PATH:
            if config.NO_UPLOAD_URL_PREFIX is None:
                raise RuntimeError(
                    "If you set NO_UPLOAD_PATH you must set NO_UPLOAD_URL_PREFIX too."
                )
        elif not config.UPLOAD_SERVICE:
            info_iq = await self.xmpp.plugin["xep_0363"].find_upload_service(
                self.infer_real_domain()
            )
            if info_iq is None:
                if self.REGISTRATION_TYPE == RegistrationType.QRCODE:
                    log.warning(
                        "No method was configured for attachment and slidge "
                        "could not automatically determine the JID of a usable upload service. "
                        "Users likely won't be able to register since this network uses a "
                        "QR-code based registration flow."
                    )
                if not config.USE_ATTACHMENT_ORIGINAL_URLS:
                    log.warning(
                        "Setting USE_ATTACHMENT_ORIGINAL_URLS to True since no method was configured "
                        "for attachments and no upload service was found. NB: this does not work for all "
                        "networks, especially for the E2EE attachments."
                    )
                    config.USE_ATTACHMENT_ORIGINAL_URLS = True
            else:
                log.info("Auto-discovered upload service: %s", info_iq["from"])
                config.UPLOAD_SERVICE = info_iq["from"]
        self.__upload_service_found.set()

    def infer_real_domain(self) -> JID:
        return JID(re.sub(r"^.*?\.", "", self.xmpp.boundjid.bare))

    async def __add_component_to_mds_whitelist(self, user_jid: JID) -> None:
        # Uses privileged entity to add ourselves to the whitelist of the PEP
        # MDS node so we receive MDS events
        iq_creation = Iq(sto=user_jid.bare, sfrom=user_jid, stype="set")
        iq_creation["pubsub"]["create"]["node"] = self["xep_0490"].stanza.NS

        try:
            await self["xep_0356"].send_privileged_iq(iq_creation)
        except PermissionError:
            log.warning(
                "IQ privileges not granted for pubsub namespace, we cannot "
                "create the MDS node of %s",
                user_jid,
            )
        except PrivilegedIqError as exc:
            nested = exc.nested_error()
            # conflict this means the node already exists, we can ignore that
            if nested is not None and nested.condition != "conflict":
                log.exception(
                    "Could not create the MDS node of %s", user_jid, exc_info=exc
                )
        except Exception as e:
            log.exception(
                "Error while trying to create to the MDS node of %s",
                user_jid,
                exc_info=e,
            )

        iq_affiliation = Iq(sto=user_jid.bare, sfrom=user_jid, stype="set")
        iq_affiliation["pubsub_owner"]["affiliations"]["node"] = self[
            "xep_0490"
        ].stanza.NS

        aff = OwnerAffiliation()
        aff["jid"] = self.boundjid.bare
        aff["affiliation"] = "member"
        iq_affiliation["pubsub_owner"]["affiliations"].append(aff)

        try:
            await self["xep_0356"].send_privileged_iq(iq_affiliation)
        except PermissionError:
            log.warning(
                "IQ privileges not granted for pubsub#owner namespace, we cannot "
                "listen to the MDS events of %s",
                user_jid,
            )
        except Exception as e:
            log.exception(
                "Error while trying to subscribe to the MDS node of %s",
                user_jid,
                exc_info=e,
            )

    @timeit
    async def login_wrap(self, session: "BaseSession") -> str:
        session.send_gateway_status("Logging in…", show="dnd")
        session.is_logging_in = True
        try:
            status = await session.login()
        except Exception as e:
            log.warning("Login problem for %s", session.user_jid, exc_info=e)
            session.send_gateway_status(f"Could not login: {e}", show="busy")
            msg = (
                "You are not connected to this gateway! "
                f"Maybe this message will tell you why: {e}"
            )
            session.send_gateway_message(msg)
            session.logged = False
            session.send_gateway_status("Login failed", show="dnd")
            return msg

        log.info("Login success for %s", session.user_jid)
        session.logged = True
        session.send_gateway_status("Syncing contacts…", show="dnd")
        with self.store.session() as orm:
            await session.contacts._fill(orm)
        if not (r := session.contacts.ready).done():
            r.set_result(True)
        if self.GROUPS:
            session.send_gateway_status("Syncing groups…", show="dnd")
            await session.bookmarks.fill()
            if not (r := session.bookmarks.ready).done():
                r.set_result(True)
        self.send_presence(pto=session.user.jid.bare, ptype="probe")
        if status is None:
            status = "Logged in"
        session.send_gateway_status(status, show="chat")

        if session.user.preferences.get("sync_avatar", False):
            session.create_task(self.fetch_user_avatar(session))
        else:
            with self.store.session(expire_on_commit=False) as orm:
                session.user.avatar_hash = None
                orm.add(session.user)
                orm.commit()
        return status

    async def fetch_user_avatar(self, session: BaseSession) -> None:
        try:
            iq = await self.xmpp.plugin["xep_0060"].get_items(
                session.user_jid.bare,
                self.xmpp.plugin["xep_0084"].stanza.MetaData.namespace,
                ifrom=self.boundjid.bare,
            )
        except IqTimeout:
            self.log.warning("Iq timeout trying to fetch user avatar")
            return
        except IqError as e:
            self.log.debug("Iq error when trying to fetch user avatar: %s", e)
            if e.condition == "item-not-found":
                try:
                    await session.on_avatar(None, None, None, None, None)
                except NotImplementedError:
                    pass
                else:
                    with self.store.session(expire_on_commit=False) as orm:
                        session.user.avatar_hash = None
                        orm.add(session.user)
                        orm.commit()
            return
        await self.__dispatcher.on_avatar_metadata_info(
            session, iq["pubsub"]["items"]["item"]["avatar_metadata"]["info"]
        )

    def _send(
        self, stanza: MessageOrPresenceTypeVar, **send_kwargs
    ) -> MessageOrPresenceTypeVar:
        stanza.set_from(self.boundjid.bare)
        if mto := send_kwargs.get("mto"):
            stanza.set_to(mto)
        stanza.send()
        return stanza

    def raise_if_not_allowed_jid(self, jid: JID):
        if not self.jid_validator.match(jid.bare):
            raise XMPPError(
                condition="not-allowed",
                text="Your account is not allowed to use this gateway. "
                "The admin controls that with the USER_JID_VALIDATOR option.",
            )

    def send_raw(self, data: str | bytes):
        # overridden from XMLStream to strip base64-encoded data from the logs
        # to make them more readable.
        if log.isEnabledFor(level=logging.DEBUG):
            if isinstance(data, str):
                stripped = copy(data)
            else:
                stripped = data.decode("utf-8")
            # there is probably a way to do that in a single RE,
            # but since it's only for debugging, the perf penalty
            # does not matter much
            for el in LOG_STRIP_ELEMENTS:
                stripped = re.sub(
                    f"(<{el}.*?>)(.*)(</{el}>)",
                    "\1[STRIPPED]\3",
                    stripped,
                    flags=re.DOTALL | re.IGNORECASE,
                )
            log.debug("SEND: %s", stripped)
        if not self.transport:
            raise NotConnectedError()
        if isinstance(data, str):
            data = data.encode("utf-8")
        self.transport.write(data)

    def get_session_from_jid(self, j: JID):
        try:
            return self._session_cls.from_jid(j)
        except XMPPError:
            pass

    def exception(self, exception: Exception) -> None:
        # """
        # Called when a task created by slixmpp's internal (eg, on slix events) raises an Exception.
        #
        # Stop the event loop and exit on unhandled exception.
        #
        # The default :class:`slixmpp.basexmpp.BaseXMPP` behaviour is just to
        # log the exception, but we want to avoid undefined behaviour.
        #
        # :param exception: An unhandled :class:`Exception` object.
        # """
        if isinstance(exception, IqError):
            iq = exception.iq
            log.error("%s: %s", iq["error"]["condition"], iq["error"]["text"])
            log.warning("You should catch IqError exceptions")
        elif isinstance(exception, IqTimeout):
            iq = exception.iq
            log.error("Request timed out: %s", iq)
            log.warning("You should catch IqTimeout exceptions")
        elif isinstance(exception, SyntaxError):
            # Hide stream parsing errors that occur when the
            # stream is disconnected (they've been handled, we
            # don't need to make a mess in the logs).
            pass
        else:
            if exception:
                log.exception(exception)
            self.loop.stop()
            exit(1)

    async def make_registration_form(
        self, _jid: JID, _node: str, _ifrom: JID, iq: Iq
    ) -> Form:
        self.raise_if_not_allowed_jid(iq.get_from())
        reg = iq["register"]
        with self.store.session() as orm:
            user = (
                orm.query(GatewayUser).filter_by(jid=iq.get_from().bare).one_or_none()
            )
        log.debug("User found: %s", user)

        form = reg["form"]
        form.add_field(
            "FORM_TYPE",
            ftype="hidden",
            value="jabber:iq:register",
        )
        form["title"] = f"Registration to '{self.COMPONENT_NAME}'"
        form["instructions"] = self.REGISTRATION_INSTRUCTIONS

        if user is not None:
            reg["registered"] = False
            form.add_field(
                "remove",
                label="Remove my registration",
                required=True,
                ftype="boolean",
                value=False,
            )

        for field in self.REGISTRATION_FIELDS:
            if field.var in reg.interfaces:
                val = None if user is None else user.get(field.var)
                if val is None:
                    reg.add_field(field.var)
                else:
                    reg[field.var] = val

        reg["instructions"] = self.REGISTRATION_INSTRUCTIONS

        for field in self.REGISTRATION_FIELDS:
            form.add_field(
                field.var,
                label=field.label,
                required=field.required,
                ftype=field.type,
                options=field.options,
                value=field.value if user is None else user.get(field.var, field.value),
            )

        reply = iq.reply()
        reply.set_payload(reg)
        return reply

    async def user_prevalidate(
        self, ifrom: JID, form_dict: dict[str, str | None]
    ) -> JSONSerializable | None:
        # Pre validate a registration form using the content of self.REGISTRATION_FIELDS
        # before passing it to the plugin custom validation logic.
        for field in self.REGISTRATION_FIELDS:
            if field.required and not form_dict.get(field.var):
                raise ValueError(f"Missing field: '{field.label}'")

        return await self.validate(ifrom, form_dict)

    @abc.abstractmethod
    async def validate(
        self, user_jid: JID, registration_form: dict[str, str | None]
    ) -> JSONSerializable | None:
        """
        Validate a user's initial registration form.

        Should raise the appropriate :class:`slixmpp.exceptions.XMPPError`
        if the registration does not allow to continue the registration process.

        If :py:attr:`REGISTRATION_TYPE` is a
        :attr:`.RegistrationType.SINGLE_STEP_FORM`,
        this method should raise something if it wasn't possible to successfully
        log in to the legacy service with the registration form content.

        It is also used for other types of :py:attr:`REGISTRATION_TYPE` too, since
        the first step is always a form. If :attr:`.REGISTRATION_FIELDS` is an
        empty list (ie, it declares no :class:`.FormField`), the "form" is
        effectively a confirmation dialog displaying
        :attr:`.REGISTRATION_INSTRUCTIONS`.

        :param user_jid: JID of the user that has just registered
        :param registration_form: A dict where keys are the :attr:`.FormField.var` attributes
            of the :attr:`.BaseGateway.REGISTRATION_FIELDS` iterable.
            This dict can be modified and will be accessible as the ``legacy_module_data``
            of the

        :return : A dict that will be stored as the persistent "legacy_module_data"
            for this user. If you don't return anything here, the whole registration_form
            content will be stored.
        """
        raise NotImplementedError

    async def validate_two_factor_code(
        self, user: GatewayUser, code: str
    ) -> JSONSerializable | None:
        """
        Called when the user enters their 2FA code.

        Should raise the appropriate :class:`slixmpp.exceptions.XMPPError`
        if the login fails, and return successfully otherwise.

        Only used when :attr:`REGISTRATION_TYPE` is
        :attr:`.RegistrationType.TWO_FACTOR_CODE`.

        :param user: The :class:`.GatewayUser` whose registration is pending
            Use their :attr:`.GatewayUser.bare_jid` and/or
            :attr:`.registration_form` attributes to get what you need.
        :param code: The code they entered, either via "chatbot" message or
            adhoc command

        :return : A dict which keys and values will be added to the persistent "legacy_module_data"
            for this user.
        """
        raise NotImplementedError

    async def get_qr_text(self, user: GatewayUser) -> str:
        """
        This is where slidge gets the QR code content for the QR-based
        registration process. It will turn it into a QR code image and send it
        to the not-yet-fully-registered :class:`.GatewayUser`.

        Only used in when :attr:`BaseGateway.REGISTRATION_TYPE` is
        :attr:`.RegistrationType.QRCODE`.

        :param user: The :class:`.GatewayUser` whose registration is pending
            Use their :attr:`.GatewayUser.bare_jid` and/or
            :attr:`.registration_form` attributes to get what you need.
        """
        raise NotImplementedError

    async def confirm_qr(
        self,
        user_bare_jid: str,
        exception: Exception | None = None,
        legacy_data: JSONSerializable | None = None,
    ) -> None:
        """
        This method is meant to be called to finalize QR code-based registration
        flows, once the legacy service confirms the QR flashing.

        Only used in when :attr:`BaseGateway.REGISTRATION_TYPE` is
        :attr:`.RegistrationType.QRCODE`.

        :param user_bare_jid: The bare JID of the almost-registered
            :class:`GatewayUser` instance
        :param exception: Optionally, an XMPPError to be raised to **not** confirm
            QR code flashing.
        :param legacy_data: dict which keys and values will be added to the persistent
            "legacy_module_data" for this user.
        """
        fut = self.qr_pending_registrations[user_bare_jid]
        if exception is None:
            fut.set_result(legacy_data)
        else:
            fut.set_exception(exception)

    async def unregister_user(
        self, user: GatewayUser, msg: str = "You unregistered from this gateway."
    ) -> None:
        self.send_presence(pshow="dnd", pstatus=msg, pto=user.jid)
        await self.xmpp.plugin["xep_0077"].api["user_remove"](None, None, user.jid)
        await self.xmpp._session_cls.kill_by_jid(user.jid)

    async def unregister(self, session: BaseSession) -> None:
        """
        Optionally override this if you need to clean additional
        stuff after a user has been removed from the persistent user store.

        By default, this just calls :meth:`BaseSession.logout`.

        :param session: The session of the user who just unregistered
        """
        try:
            await session.logout()
        except NotImplementedError:
            pass

    async def input(
        self,
        jid: JID,
        text: str | None = None,
        mtype: MessageTypes = "chat",
        **input_kwargs: Any,
    ) -> str:
        """
        Request arbitrary user input using a simple chat message, and await the result.

        You shouldn't need to call this directly bust instead use
        :meth:`.BaseSession.input` to directly target a user.

        :param jid: The JID we want input from
        :param text: A prompt to display for the user
        :param mtype: Message type
        :return: The user's reply
        """
        return await self.__chat_commands_handler.input(
            jid, text, mtype=mtype, **input_kwargs
        )

    async def send_qr(self, text: str, **msg_kwargs: Any) -> None:
        """
        Sends a QR Code to a JID

        You shouldn't need to call directly bust instead use
        :meth:`.BaseSession.send_qr` to directly target a user.

        :param text: The text that will be converted to a QR Code
        :param msg_kwargs: Optional additional arguments to pass to
            :meth:`.BaseGateway.send_file`, such as the recipient of the QR,
            code
        """
        qr = qrcode.make(text)
        with tempfile.NamedTemporaryFile(
            suffix=".png", delete=config.NO_UPLOAD_METHOD != "move"
        ) as f:
            qr.save(f.name)
            await self.send_file(Path(f.name), **msg_kwargs)

    def shutdown(self) -> list[asyncio.Task[None]]:
        # """
        # Called by the slidge entrypoint on normal exit.
        #
        # Sends offline presences from all contacts of all user sessions and from
        # the gateway component itself.
        # No need to call this manually, :func:`slidge.__main__.main` should take care of it.
        # """
        log.debug("Shutting down")
        tasks = []
        with self.store.session() as orm:
            for user in orm.query(GatewayUser).all():
                tasks.append(self._session_cls.from_jid(user.jid).shutdown())
                self.send_presence(ptype="unavailable", pto=user.jid)
        return tasks


SLIXMPP_PLUGINS = [
    "link_preview",  # https://wiki.soprani.ca/CheogramApp/LinkPreviews
    "xep_0030",  # Service discovery
    "xep_0045",  # Multi-User Chat
    "xep_0050",  # Adhoc commands
    "xep_0054",  # VCard-temp (for MUC avatars)
    "xep_0055",  # Jabber search
    "xep_0059",  # Result Set Management
    "xep_0066",  # Out of Band Data
    "xep_0071",  # XHTML-IM (for stickers and custom emojis maybe later)
    "xep_0077",  # In-band registration
    "xep_0084",  # User Avatar
    "xep_0085",  # Chat state notifications
    "xep_0100",  # Gateway interaction
    "xep_0106",  # JID Escaping
    "xep_0115",  # Entity capabilities
    "xep_0122",  # Data Forms Validation
    "xep_0128",  # Service Discovery Extensions
    "xep_0153",  # vCard-Based Avatars (for MUC avatars)
    "xep_0172",  # User nickname
    "xep_0184",  # Message Delivery Receipts
    "xep_0199",  # XMPP Ping
    "xep_0221",  # Data Forms Media Element
    "xep_0231",  # Bits of Binary (for stickers and custom emojis maybe later)
    "xep_0249",  # Direct MUC Invitations
    "xep_0264",  # Jingle Content Thumbnails
    "xep_0280",  # Carbons
    "xep_0292_provider",  # VCard4
    "xep_0308",  # Last message correction
    "xep_0313",  # Message Archive Management
    "xep_0317",  # Hats
    "xep_0319",  # Last User Interaction in Presence
    "xep_0333",  # Chat markers
    "xep_0334",  # Message Processing Hints
    "xep_0356",  # Privileged Entity
    "xep_0363",  # HTTP file upload
    "xep_0385",  # Stateless in-line media sharing
    "xep_0402",  # PEP Native Bookmarks
    "xep_0421",  # Anonymous unique occupant identifiers for MUCs
    "xep_0424",  # Message retraction
    "xep_0425",  # Message moderation
    "xep_0444",  # Message reactions
    "xep_0447",  # Stateless File Sharing
    "xep_0461",  # Message replies
    "xep_0469",  # Bookmark Pinning
    "xep_0490",  # Message Displayed Synchronization
    "xep_0492",  # Chat Notification Settings
]

LOG_STRIP_ELEMENTS = ["data", "binval"]

log = logging.getLogger(__name__)
